#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


#pylint:disable=undefined-all-variable

from .paths import (
    list_images,
    list_videos,
    listdir,
    is_image,
    is_video,
    cwd,
    exists,
    get_size,
    abspath,
    isdir,
    isfile,
    mkdir,
    osname,
    chdir,
    join,
    dirname,

    # Variables
    _acceptable_video_formats,
    _acceptable_image_formats,

    __all__ as __all_paths__,
)

__all__ = __all_paths__ 

# Stop polluting the namespace
del __all_paths__


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


from .mean_subtraction import MeanProcess
from .mean_subtraction import compute_mean
from .mean_subtraction import compute_mean_from_dir
from .mean_subtraction import subtract_mean
from .mean_subtraction import _check_mean_sub_values


# __all__ globals 
from .mean_subtraction import __all__ as __all_mean__ 

__all__ = __all_mean__

del __all_mean__



#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

# pylint:disable=unused-argument

import numpy as np

from ._tensor_str import _str
from ._tensor_base import _TensorBase


# We use np.ndarray as a super class, because ``ndarray.view()`` expects an ndarray sub-class
# We also derive useful class methods from ``_TensorBase`` which serves as the Tensor's Base
class Tensor(_TensorBase, np.ndarray):
    # This is required to get the type(Tensor) to be 'caer.Tensor'.
    # Without this, type(Tensor) is 'caer.tensor.Tensor' which is not ideal.
    # Alternatively, we may shove this class to __init__.py, but this would, again, not be ideal
    __module__ = 'caer'

    # def __new__(cls, x, dtype=None):
    def __new__(cls, x, cspace, dtype=None):
        if not isinstance(x, (tuple, list, np.ndarray)):
            raise TypeError('Data needs to be (ideally) a list')

        if not isinstance(cspace, str) and cspace is not None:
            raise TypeError(
                f'`cspace` needs to be of type <string>, not {type(cspace)}'
            )

        obj = np.asarray(x, dtype=dtype).view(cls)
        obj.dtype = obj.dtype

        return obj

    def __init__(self, x, cspace, dtype=None):
        super().__init__()  # gets attributes from '_TensorBase'
        self.x = x

        if cspace is None:
            self.cspace = 'null'

        else:
            if cspace in ('rgb', 'bgr', 'gray', 'hsv', 'hls', 'lab', 'yuv', 'luv'):
                self.cspace = cspace
            else:
                raise ValueError(
                    'The `cspace` attribute needs to be either rgb/bgr/gray/hsv/hls/lab/yuv/luv'
                )

    def __repr__(self):
        return _str(self)

    def __str__(self):
        return self.__repr__()


def is_tensor(x):
    r'''
        Returns True if `x` is a Caer tensor.

        Note that this function is simply doing ``isinstance(x, Tensor)``. Using the ``isinstance`` check is better for typechecking with mypy, and more explicit - so it's recommended to use that instead of ``is_tensor``.

        For now, Caer Tensors are simply customized Numpy arrays.

    Args:
        x (Object): Object to test
    '''
    return 'caer.Tensor' in str(type(x)) or (isinstance(x, Tensor))



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import math 
import cv2 as cv

from ..coreten import Tensor, to_tensor
from .._internal import _check_target_size
from ..globals import (
    INTER_AREA, INTER_CUBIC, INTER_NEAREST, INTER_LINEAR
)

__all__ = [
    'resize'
]


def resize(tens, target_size=None, resize_factor=None, preserve_aspect_ratio=False, interpolation='bilinear'):
    r"""
        Resizes an image to a target_size without aspect ratio distortion.
        
        Your output images will be of size ``target_size``, and will not be distorted. Instead, the parts of the image that do not fit within the target size get cropped out.

        The resizing process is:
        1. Resize the image as minimally as possible.
        2. Take the largest centered crop of the image with dimensions = ``target_size``.
        
        Alternatively, you may use:
        ```python
        size = (200,200)
        tens = caer.resize(tens, target_size=size, preserve_aspect_ratio=True)
        ``` 

        Note:
            ``caer.imread()`` comes with an in-built functionality to resize your images, eliminating the need for you to call ``caer.resize()``. This is purely optional and may appeal to certain users. 

            You may also use ``caer.smart_resize()`` for on-the-fly image resizing that `preserves the aspect ratio`. 

    
        Args:
            tens (Tensor): Input Image. Must be in the format ``(height, width, channels)``.
            target_size (tuple): Target size. Must be a tuple of ``(width, height)`` integer.
            resize_factor (float, tuple): Resizing Factor to employ. 
                Shrinks the image if ``resize_factor < 1``
                Enlarges the image if ``resize_factor > 1``
            preserve_aspect_ratio (bool): Prevent aspect ratio distortion (employs center crop).
            interpolation (str): Interpolation to use for resizing. Defaults to `'bilinear'`. 
                Supports `'bilinear'`, `'bicubic'`, `'area'`, `'nearest'`.
        
        
        Returns:
            Tensor of shape ``(height, width, channels)``.


        Examples::

            >> tens = caer.data.sunrise()
            >> tens.shape
            (427, 640, 3)

            >> resized = caer.resize(tens, target_size=(200,200)) # Hard-resize. May distort aspect ratio
            >> resized.shape
            (200, 200, 3)

            >> resized_wf = caer.resize(tens, resize_factor=.5) # Resizes the image to half its dimensions
            >> resized_wf.shape
            (213, 320, 3)

            >> resized = caer.resize(tens, target_size=(200,200), preserve_aspect_ratio=True) # Preserves aspect ratio
            >> resized.shape
            (200, 200, 3)

    """
    # Opencv uses the (h,w) format
    height, width = tens.shape[:2]
    interpolation = str(interpolation)
    cspace = None 

    if isinstance(tens, Tensor):
        # We'll need to preserve this before returning
        cspace = tens.cspace 


    if resize_factor is None:
        if target_size is None:
            if preserve_aspect_ratio:
                raise ValueError('Specify a target size')
            else:
                raise ValueError('Specify either a resize factor or target dimensions')
        if target_size is not None:
            if len(target_size) == 2:
                new_shape = target_size
            else:
                raise ValueError('Tuple shape must be = 2 (width, height)')

    if resize_factor is not None:
        target_size = None
        preserve_aspect_ratio = False 

        if not isinstance(resize_factor, (int, float)):
            raise ValueError('resize_factor must be an integer or float')

        if resize_factor > 1:
            interpolation = 'bicubic'

        new_shape = (int(resize_factor * width), int(resize_factor * height))
            
    interpolation_methods = {
        'nearest': INTER_NEAREST, '0': INTER_NEAREST, 0: INTER_NEAREST, # 0
        'bilinear': INTER_LINEAR, '1': INTER_LINEAR,  1: INTER_LINEAR,  # 1
        'bicubic': INTER_CUBIC,   '2': INTER_CUBIC,   2: INTER_CUBIC,   # 2
        'area': INTER_AREA,       '3': INTER_AREA,    3: INTER_AREA     # 3
    }

    if interpolation not in interpolation_methods:
        raise ValueError('Specify a valid interpolation type - area/nearest/bicubic/bilinear')

    if preserve_aspect_ratio:
        im = _resize_with_ratio(tens, target_size=target_size, preserve_aspect_ratio=preserve_aspect_ratio, interpolation=interpolation_methods[interpolation])
    else:
        width, height = new_shape[:2]
        im = _cv2_resize(tens, (width, height), interpolation=interpolation_methods[interpolation])
    
    # For this function, the <cspace> attribute is not required.
    # So, we disable the mandatory check that the <cspace> attribute needs to be passed for 
    # foreign Tensors/ndarrays 
    return to_tensor(im, cspace=cspace, override_checks=True)


def smart_resize(tens, target_size, interpolation='bilinear'):
    r"""
        Resizes an image to a target_size without aspect ratio distortion.
        
        Your output images will be of size `target_size`, and will not be distorted. Instead, the parts of the image that do not fit within the target size get cropped out.

        The resizing process is:
        1. Resize the image as minimally as possible.
        2. Take the largest centered crop of the image with dimensions = `target_size`.
        
        Alternatively, you may use:
        ```python
        size = (200,200)
        tens = caer.resize(tens, target_size=size, preserve_aspect_ratio=True)
        ``` 
    
        Args:
            tens (Tensor): Input Image. Must be in the format `(height, width, channels)`.
            target_size (tuple): Target size. Must be a tuple of `(width, height)` integer.
            interpolation (str): Interpolation to use for resizing. Defaults to `'bilinear'`. 
                Supports `'bilinear'`, `'bicubic'`, `'area'`, `'nearest'`.
        
        Returns:
            Tensor of shape `(height, width, channels)`

        Examples::
        
            >> tens = caer.data.sunrise()
            >> tens.shape
            (427, 640, 3)
            >> resized = caer.smart_resize(tens, target_size=(200,200))
            >> resized.shape
            (200, 200, 3)

    """
    # if not isinstance(tens, Tensor):
    #     raise ValueError('To use `caer.smart_resize()`, `tens` needs to be a caer.Tensor')

    im = _resize_with_ratio(tens, target_size=target_size, preserve_aspect_ratio=True, interpolation=interpolation)

    # For this function, the <cspace> attribute is not required.
    # So, we disable the mandatory check that the <cspace> attribute needs to be passed for 
    # foreign Tensors/ndarrays 
    return to_tensor(im, override_checks=True)


def _cv2_resize(image, target_size, interpolation=None):
    """
    ONLY TO BE USED INTERNALLY. NOT AVAILABLE FOR EXTERNAL USAGE. 
    Resizes the image ignoring the aspect ratio of the original image
    """
    _ = _check_target_size(target_size)

    width, height = target_size[:2]

    if interpolation is None:
        interpolation = INTER_AREA

    dimensions = (width, height)

    return cv.resize(image, dimensions, interpolation=interpolation)


def _resize_with_ratio(tens, target_size, preserve_aspect_ratio=False, interpolation='bilinear'):
    """
        Resizes an image using advanced algorithms
        :param target_size: Tuple of size 2 in the format (width,height)
        :param preserve_aspect_ratio: Boolean to keep/ignore aspect ratio when resizing
    """
    _ = _check_target_size(target_size)
    interpolation = str(interpolation)
    
    if not isinstance(preserve_aspect_ratio, bool):
        raise ValueError('preserve_aspect_ratio must be a boolean')
    
    interpolation_methods = {
        'nearest': INTER_NEAREST, '0': INTER_NEAREST, # 0
        'bilinear': INTER_LINEAR, '1': INTER_LINEAR,# 1
        'bicubic': INTER_CUBIC, '2': INTER_CUBIC,# 2
        'area': INTER_AREA, '3': INTER_AREA,# 3
    }

    if interpolation not in interpolation_methods:
        raise ValueError('Specify a valid interpolation type - area/nearest/bicubic/bilinear')

    oh, ow = tens.shape[:2]
    target_w, target_h = target_size

    if target_h > oh or target_w > ow:
        raise ValueError('To compute resizing keeping the aspect ratio, the target size dimensions must be <= actual image dimensions')

    # Computing minimal resize
    # min_width, w_factor = _compute_minimal_resize(ow, target_w)
    # min_height, h_factor = _compute_minimal_resize(oh, target_h)
    minimal_resize_factor = _compute_minimal_resize((ow, oh), (target_w, target_h))

    # Resizing minimally
    tens = _cv2_resize(tens, (ow//minimal_resize_factor, oh//minimal_resize_factor))

    # Computing centre crop (to avoid extra crop, we resize minimally first)
    tens = _compute_centre_crop(tens, (target_w, target_h))

    if tens.shape[:2] != target_size[:2]:
        tens = _cv2_resize(tens, (target_w, target_h), interpolation=interpolation_methods[interpolation])
    
    return tens
    

def _compute_minimal_resize(org_size, target_dim):
    # for i in range(10):
    #     i += 1
    #     d = dim*i
    #     if org_dim >= d and dim < dim*(i+1):
    #         if (org_dim - dim*(i+1)) > dim:
    #             continue
    #         else:
    #             return d, i
    # import math 
    # mi = math.floor(org_dim/dim)
    # d = dim * mi 
    # return d, mi

    if not isinstance(org_size, tuple) or not isinstance(target_dim, tuple):
        raise ValueError('org_size and target_dim must be a tuple')

    if len(org_size) != 2 or len(target_dim) != 2:
        raise ValueError('Size of tuple must be = 2')

    oh, ow = org_size[:2]
    targ_w, targ_h = target_dim[:2]

    h_factor = math.floor(oh/targ_h)
    w_factor = math.floor(ow/targ_w)

    if h_factor <= w_factor:
        return h_factor 
    else:
        return w_factor


def _compute_centre_crop(tens, target_size):
    _ = _check_target_size(target_size)

    # Getting org height and target
    oh, ow = tens.shape[:2]
    target_w, target_h = target_size

    # The following line is actually the right way of accessing height and width of an opencv-specific image (height, width). However for some reason, while the code runs, this is flipped (it now becomes (width,height)). Testing needs to be done to catch this little bug
    # oh, ow = tens.shape[:2]


    if target_h > oh or target_w > ow:
        raise ValueError('To compute centre crop, target size dimensions must be <= tens dimensions')

    diff_h = (oh - target_h) // 2
    diff_w = (ow - target_w ) // 2
    
    # tens[y:y+h, x:x+h]
    return tens[diff_h:diff_h + target_h, diff_w:diff_w + target_w]


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


from threading import Thread
import time
import math
from queue import Queue
import cv2 as cv

from .constants import FRAME_COUNT, FPS

__all__ = [
    'GPUFileStream'
]


class GPUFileStream:
    r"""
        This is an auxiliary class that enables Video Streaming using the GPU for caer with minimalistic latency, and at the expense of little to no additional computational requirements.
        
        The basic idea behind it is to tracks and save the salient feature array for the given number of frames and then uses these anchor point to cancel out all perturbations relative to it for the incoming frames in the queue. This class relies heavily on **Threaded Queue mode** for error-free & ultra-fast frame handling.

    Args:
        source (int, str): Source path for the video. If ``source=0``, the default camera device is used. For 
            multiple external camera devices, use incremented values. For eg: ``source=1`` represents the second camera device on your system.
        qsize (int): Default queue size for handling the video streams. Default: 128.
    """

    def __init__(self, source, qsize=128):
        """
            Source must be a path to a video file
            Utilizes your system's GPU to process the stream
        """

        if not isinstance(source, str):
            raise ValueError(f'Expected either a filepath. Got {type(source)}. Consider using VideoStream which supports both live video as well as pre-existing videos')

        # initialize the file video stream along with the boolean
        # used to indicate if the thread should be stopped or not
        self.stream = cv.VideoCapture(source)
        self.kill_stream = False
        self.count = 0

        # initialize the queue to store frames
        self.Q = Queue(maxsize=qsize)

        self.width = int(self.stream.get(cv.CAP_PROP_FRAME_WIDTH))
        self.height = int(self.stream.get(cv.CAP_PROP_FRAME_HEIGHT))
        self.res = (self.width, self.height)

        self.fps = math.ceil(self.stream.get(FPS))
        self.frames = int(self.stream.get(FRAME_COUNT))
        
        # since we use UMat to store the images to
        # we need to initialize them beforehand
        self.qframes = [0] * qsize
        for ii in range(qsize):
            self.qframes[ii] = cv.UMat(self.height, self.width, cv.CV_8UC3)


    def begin_stream(self):
        # start a thread to read frames from the file video stream
        t = Thread(target=self.update, args=())
        t.daemon = True
        t.start()
        return self


    def update(self):
        # keep looping infinitely
        while True:
            if self.kill_stream:
                return

            # otherwise, ensure the queue has room in it
            if not self.Q.full():
                self.count += 1
                target = (self.count-1) % self.Q.maxsize
                ret = self.stream.grab()

                if not ret:
                    self.release()
                    return 

                self.stream.retrieve(self.qframes[target])

                # add the frame to the queue
                self.Q.put(target)


    def read(self):
        while (not self.more() and self.kill_stream):
            time.sleep(0.1)
        # return next frame in the queue
        return self.qframes[self.Q.get()]


    def more(self):
        # return True if there are still frames in the queue
        return self.Q.qsize() > 0


    def release(self):
        self.kill_stream = True
        # wait until stream resources are released
        self.thread.join()


    # Gets frame count
    def count_frames(self):
        if not self.kill_stream and not self.live_video:
            return self.frames
            # if get_opencv_version() == '2':
            #     return int(self.stream.get(FRAME_COUNT_DEPR))
            # else:
            #     return int(self.stream.get(FRAME_COUNT))
            

        if self.live_video:
            print('[WARNING] Frames cannot be computed on live streams')
            return -1


    # Gets FPS count
    def get_fps(self):
        if not self.kill_stream:
            return self.fps

    # Get frame dimensions
    def get_res(self):
        return self.res


#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


## Color Spaces
IMREAD_COLOR = 1
BGR2RGB = 4
BGR2GRAY = 6
RGB2GRAY = 7
GRAY2BGR = 8
GRAY2RGB = 8
BGR2HSV = 40
RGB2BGR = BGR2RGB
RGB2HSV = 41
BGR2LAB = 44
RGB2LAB = 45
BGR2LUV = 50
RGB2LUV = 51
BGR2HLS = 52
RGB2HLS = 53
HSV2BGR = 54
HSV2RGB = 55
LAB2BGR = 56
LAB2RGB = 57
LUV2BGR = 58
LUV2RGB = 59
HLS2BGR = 60
HLS2RGB = 61
BGR2YUV = 82
RGB2YUV = 83
YUV2BGR = 84
YUV2RGB = 85
__all__ = [d for d in dir() if not d.startswith('_')]



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import os
import random
import time
import numpy as np

from .io import imread
from ._internal import _check_target_size
from .path import join, list_images
from .color import to_gray

__all__ = [
    'preprocess_from_dir',
    'sep_train',
    'normalize',
    'train_val_split'
]


def preprocess_from_dir(DIR, 
                        classes=None, 
                        IMG_SIZE=None, 
                        channels=3, 
                        isShuffle=True, 
                        save_data=False, 
                        destination_filename=None, 
                        verbose=True):
    """
    Reads Images in base directory DIR using 'classes' (computed from sub directories )
    Arguments:
        DIR (str): Base directory 
        classes (list): A list of folder names within `DIR`.
        IMG_SIZE (tuple): Image Size tuple of size 2 (width, height)
        channels (int): Number of channels each image will be processed to (default: 3)
        isShuffle (bool): Shuffle the training set
        save_data (bool): If True, saves the training set as a .npy or .npz file based on destination_filename
        destination_filename (str): if save_data is True, the train set will be saved as the filename specified
        verbose (bool): Displays the progress to the terminal as preprocessing continues. Default = True
    
    Returns
        data: Image Pixel Values with corresponding labels (float32)
        Saves the above variables as .npy files if `save_data = True`
    """
    data = [] 

    if not os.path.exists(DIR):
        raise ValueError('The specified directory does not exist')

    if IMG_SIZE is None:
        raise ValueError('IMG_SIZE must be specified')

    if not isinstance(IMG_SIZE, tuple) or len(IMG_SIZE) != 2:
        raise ValueError('IMG_SIZE must be a tuple of size 2 (width,height)')

    if not isinstance(save_data, bool):
        raise ValueError('save_data must be a boolean (True/False)')

    if not isinstance(classes, list):
        raise ValueError('"classes" must be a list')

    if save_data:
        if destination_filename is None:
            raise ValueError('Specify a destination file name')

        elif not ('.npy' in destination_filename or '.npz' in destination_filename):
            raise ValueError('Specify the correct numpy destination file extension (.npy or .npz)')
    
    if not save_data and destination_filename is not None:
        destination_filename = None

    # Loading from Numpy Files
    if destination_filename is not None and os.path.exists(destination_filename):
        print('[INFO] Loading from Numpy Files')
        since = time.time()
        data = np.load(destination_filename, allow_pickle=True)
        end = time.time()
        took = end - since
        print('----------------------------------------------')
        print(f'[INFO] Loaded in {took:.0f}s from Numpy Files')

        return data

    # Extracting image data and adding to `data`
    else:
        if destination_filename is not None:
            print(f'[INFO] Could not find {destination_filename}. Generating the training data')
        else:
            print('[INFO] Could not find a file to load from. Generating the training data')
        print('----------------------------------------------')

        # Starting timer
        since_preprocess = time.time()

        for item in classes:
            class_path = join(DIR, item)
            class_label = classes.index(item)
            count = 0 
            tens_list = list_images(class_path, use_fullpath=True, verbose=verbose)

            for image_path in tens_list:
                tens = imread(image_path, target_size=IMG_SIZE, rgb=True)

                if tens is None:
                    continue
                
                # Gray
                if channels == 1:
                    tens = to_gray(tens)

                # Appending to train set
                data.append([tens, class_label])
                count += 1

        # Shuffling the Training Set
        if isShuffle is True:
            random.shuffle(data)

        # Converting to Numpy
        data = np.array(data)

        # Saves the Data set as a .npy file
        if save_data:
            #Converts to Numpy and saves
            if destination_filename.endswith('.npy'):
                print('[INFO] Saving as .npy file')
            elif destination_filename.endswith('.npz'):
                print('[INFO] Saving as .npz file')
            
            # Saving
            since = time.time()
            np.save(destination_filename, data)
            end = time.time()
            
            time_elapsed = end-since
            minu_elapsed = time_elapsed // 60
            sec_elapsed = time_elapsed % 60
            print(f'[INFO] {destination_filename} saved! Took {minu_elapsed:.0f}m {sec_elapsed:.0f}s')

        #Returns Training Set
        end_preprocess = time.time()
        time_elapsed_preprocess = end_preprocess - since_preprocess
        minu = time_elapsed_preprocess // 60
        sec = time_elapsed_preprocess % 60

        print('----------------------------------------------')
        print(f'[INFO] {len(data)} files preprocessed! Took {minu:.0f}m {sec:.0f}s')

        return data

def normalize(x, dtype='float32'):
    """
    Normalizes the data to mean 0 and standard deviation 1
    """
    # x/=255.0 raises a TypeError
    # x = x/255.0
    
    # Converting to float32 and normalizing (float32 saves memory)
    x = x.astype(dtype) / 255
    return x

def sep_train(data, IMG_SIZE, channels=1):
    # x = []
    # y = []
    # for feature, label in data:
    #     x.append(feature)
    #     y.append(label)
    
    _ = _check_target_size(IMG_SIZE)

    x = [i[0] for i in data]
    y = [i[1] for i in data]

    # Without reshaping, X.shape --> (no. of images, IMG_SIZE, IMG_SIZE)
    # On reshaping, X.shape --> (no. of images, IMG_SIZE, IMG_SIZE, channels)

    # Converting to Numpy + Reshaping X
    x = reshape(x, IMG_SIZE, channels)
    y = np.array(y)

    return x, y


def reshape(x, IMG_SIZE, channels):
    _ = _check_target_size(IMG_SIZE)

    width, height = IMG_SIZE[:2]
    return np.array(x).reshape(-1, width, height, channels)


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

#pylint:disable=unused-variable

import cv2 as cv 
import random 
import math 
import numpy as np 

from ..coreten import Tensor, to_tensor

from ..color import (
    to_hls, 
    to_rgb,
    to_bgr,
    to_hsv,
    to_gray,
    to_lab
)

def _hls(tens) -> Tensor:
    tens = to_tensor(tens, enforce_tensor=True)
    return to_hls(tens)


def _bgr(tens) -> Tensor:
    tens = to_tensor(tens, enforce_tensor=True)
    return to_bgr(tens)


def _rgb(tens) -> Tensor:
    tens = to_tensor(tens, enforce_tensor=True)
    return to_rgb(tens)


def _hue(tens) -> Tensor:
    return _hls(tens)[:,:,0]


def _get_tens_size(tens) -> Tensor:
    r"""
        Returns image size as (width, height)
    """
    h, w = tens.shape[:2]

    return (w, h)


def _get_num_channels(tens) -> int:
    r"""
        We assume only images of 1 and 3 channels
    """
    if len(tens.shape) == 3 and tens.shape[2] == 3:
        return 3
    
    else:
        return 1


def is_tuple(x) -> bool:
    return isinstance(x, tuple)


def is_list(x):
    return isinstance(x, list)


def is_numeric(x):
    return isinstance(x, int)


def is_numeric_list_or_tuple(x):
    for i in x:
        if not is_numeric(i):
            return False
    return True


def _snow_process(tens, snow_coeff) -> Tensor:
    tens = _hls(tens)
    cspace = tens.cspace 

    tens = np.array(tens, dtype=np.float64) 

    brightness_coefficient = 2.5 

    ## scale pixel values up for channel 1 (lightness)
    tens[:,:,1][tens[:,:,1]<snow_coeff] = tens[:,:,1][tens[:,:,1]<snow_coeff]*brightness_coefficient 
    tens[:,:,1][tens[:,:,1]>255]  = 255 ##Sets all values above 255 to 255

    tens = np.array(tens, dtype=np.uint8)

    return to_tensor(tens, cspace=cspace)


def _generate_random_lines(imshape, slant, drop_length, rain_type):
    drops=[]
    area=imshape[0] * imshape[1]
    num_drops = area//600

    if rain_type.lower() == 'drizzle':
        num_drops = area//770
        drop_length = 10

    elif rain_type.lower() == 'heavy':
        drop_length = 30

    elif rain_type.lower() == 'torrential':
        num_drops = area//500
        drop_length = 60

    for i in range(num_drops) : ## If you need heavier rain, try increasing this
        if slant < 0:
            x = np.random.randint(slant, imshape[1])
        else:
            x = np.random.randint(0, imshape[1] - slant)

        y = np.random.randint(0, imshape[0] - drop_length)
        drops.append((x,y))

    return drops, drop_length


def _rain_process(tens, slant, drop_length, drop_color, drop_width, rain_drops) -> Tensor:
    tens = to_tensor(tens, enforce_tensor=True)
    tens_t = tens.clone()

    for rain_drop in rain_drops:
        cv.line(tens_t, (rain_drop[0],rain_drop[1]), (rain_drop[0]+slant,rain_drop[1]+drop_length), drop_color, drop_width)

    tens = cv.blur(tens_t,(7,7)) ## Rainy views are blurred
    brightness_coefficient = 0.7 ## Rainy days are usually shady 

    tens = _hls(tens) ## Conversion to hls
    cspace = tens.cspace 

    tens[:,:,1] = tens[:,:,1]*brightness_coefficient ## scale pixel values down for channel 1 (Lightness)

    return to_tensor(tens, cspace=cspace)


def _add_blur(tens, x, y, hw, fog_coeff) -> Tensor:
    tens = to_tensor(tens, override_checks=True)

    overlay = tens.clone()
    output = tens.clone()

    alpha = 0.08 * fog_coeff
    rad = hw // 2
    point = (x+hw//2, y+hw//2)
    cv.circle(overlay, point, int(rad), (255,255,255), -1)
    cv.addWeighted(overlay, alpha, output, 1 -alpha , 0, output)

    return to_tensor(output, override_checks=True)


def _generate_random_blur_coordinates(imshape, hw):
    blur_points=[]
    midx = imshape[1]//2-2*hw
    midy = imshape[0]//2-hw
    index=1

    while midx >- hw or midy >- hw:
        for i in range(hw//10*index):
            x= np.random.randint(midx, imshape[1]-midx-hw)
            y= np.random.randint(midy, imshape[0]-midy-hw)
            blur_points.append((x,y))
        midx -= 3*hw*imshape[1]//sum(imshape)
        midy -= 3*hw*imshape[0]//sum(imshape)
        index += 1
    
    return blur_points


def _generate_gravel_patch(rectangular_roi):
    x1 = rectangular_roi[0]
    y1 = rectangular_roi[1]
    x2 = rectangular_roi[2]
    y2 = rectangular_roi[3] 
    gravels = []
    area = abs((x2-x1)*(y2-y1))

    for i in range((int(area//10))):
        x= np.random.randint(x1,x2)
        y= np.random.randint(y1,y2)
        gravels.append((x,y))

    return gravels


def _gravel_process(tens, x1, x2, y1, y2, num_patches):
    tens = to_tensor(tens, enforce_tensor=True)
    
    x = tens.shape[1]
    y = tens.shape[0]
    rectangular_roi_default = []

    for i in range(num_patches):
        xx1 = random.randint(x1, x2)
        xx2 = random.randint(x1, xx1)
        yy1 = random.randint(y1, y2)
        yy2 = random.randint(y1, yy1)
        rectangular_roi_default.append((xx2, yy2, min(xx1, xx2+200), min(yy1, yy2+30)))

    tens = _hls(tens)
    cspace = tens.cspace 

    for roi in rectangular_roi_default:
        gravels = _generate_gravel_patch(roi)
        for gravel in gravels:
            x = gravel[0]
            y = gravel[1]
            r = random.randint(1, 4)
            r1 = random.randint(0, 255)
            tens[max(y-r,0):min(y+r,y), max(x-r,0):min(x+r,x), 1] = r1

    return to_tensor(tens, cspace=cspace)


def flare_source(tens, point, radius, src_color) -> Tensor:
    r"""
        Add a source of light (flare) on an specific region of an image.

    Args:
        tens (Tensor) : Any regular BGR/RGB image.
        point (int): Starting point of the flare.
        radius (int): Intended radius (in pixels) of the flare.
        src_color (tuple): Color of the flare. Must be in the format ``(R,G,B)``
        rectangular_roi (tuple): Rectanglar co-ordinates of the intended region of interest. Default: (-1,-1,-1,-1).
        num_patches (int): Number of patches to operate on.
        rgb (bool): Operate on RGB images. Default: True.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunris)
        >> filtered = caer.filters.add_gravel(tens)
        >> filtered
        (427, 640, 3)

    """
    tens = to_tensor(tens, override_checks=True)

    if not isinstance(src_color, tuple):
        raise ValueError('`src_color` needs to be a tuple in the format `(R,G,B)`')
    
    # By default, assume a RGB-format
    # We reverse the tuple as OpenCV expects a BGR layout
    src_color = src_color[::-1]

    overlay = tens.clone()
    output = tens.clone()

    num_times = radius // 10
    alpha = np.linspace(0.0, 1, num=num_times)
    rad = np.linspace(1, radius, num=num_times)

    for i in range(num_times):
        cv.circle(overlay, point, int(rad[i]),  color=src_color, thickness=-1)
        alp = alpha[num_times-i-1]*alpha[num_times-i-1]*alpha[num_times-i-1]
        cv.addWeighted(overlay, alp, output, 1 - alp , 0, output)

    return to_tensor(tens, override_checks=True)


def _add_sun_flare_line(flare_center, angle, imshape):
    x = []
    y = []
    i = 0

    for rand_x in range(0, imshape[1], 10):
        rand_y = math.tan(angle)*(rand_x-flare_center[0])+flare_center[1]
        x.append(rand_x)
        y.append(2*flare_center[1]-rand_y)

    return x, y


def _add_sun_process(tens, num_flare_circles, flare_center, src_radius, x, y, src_color) -> Tensor:
    tens = to_tensor(tens, override_checks=True)
    overlay = tens.clone()
    output = tens.clone()
    imshape = tens.shape

    for i in range(num_flare_circles):
        alpha = random.uniform(0.05,0.2)
        r = random.randint(0, len(x)-1)
        rad = random.randint(1, imshape[0]//100-2)
        cv.circle(overlay, 
                 (int(x[r]),int(y[r])), 
                 rad*rad*rad, 
                 (random.randint(max(src_color[0]-50,0), src_color[0]),
                 random.randint(max(src_color[1]-50,0), src_color[1]), 
                 random.randint(max(src_color[2]-50,0), src_color[2])), 
                 -1
        )
        cv.addWeighted(overlay, alpha, output, 1 - alpha,0, output)            

    return flare_source(output, (int(flare_center[0]),int(flare_center[1])), src_radius, src_color)


def _apply_motion_blur(tens, count):
    tens = to_tensor(tens, override_checks=True)

    tens_t = tens.clone()
    imshape = tens_t.shape
    size = 15
    kernel_motion_blur = np.zeros((size, size))
    kernel_motion_blur[int((size-1)/2), :] = np.ones(size)
    kernel_motion_blur = kernel_motion_blur / size

    i = imshape[1]*3//4 - 10 * count

    while i <= imshape[1]:
        tens_t[:,i:,:] = cv.filter2D(tens_t[:,i:,:], -1, kernel_motion_blur)
        tens_t[:,:imshape[1]-i,:] = cv.filter2D(tens_t[:,:imshape[1]-i,:], -1, kernel_motion_blur)
        i += imshape[1]//25 - count
        count += 1

    return to_tensor(tens_t, override_checks=True)


def _autumn_process(tens) -> Tensor:
    tens = to_tensor(tens, enforce_tensor=True)
    imshape = tens.shape
    tens = _hls(tens)
    cspace = tens.cspace

    step = 8
    aut_colors=[1, 5, 9, 11]
    col= aut_colors[random.randint(0, 3)]

    for i in range(0, imshape[1], step):
        for j in range(0, imshape[0], step):
            avg = np.average(tens[j:j+step,i:i+step,0])

            if avg > 20 and avg < 100 and np.average(tens[j:j+step,i:i+step,1]) < 100:
                tens[j:j+step,i:i+step,0] = col
                tens[j:j+step,i:i+step,2] =255

    return to_tensor(tens, cspace=cspace)


def _exposure_process(tens) -> Tensor:
    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace 

    tens = to_bgr(tens)
    tens_yuv = cv.cvtColor(tens, cv.COLOR_BGR2YUV)

    clahe = cv.createCLAHE(clipLimit=2.0, tileGridSize=(4,4))
    ones = np.ones(tens_yuv[:,:,0].shape)
    ones[tens_yuv[:,:,0]>150] = 0.85
    tens_yuv[:,:,0] = tens_yuv[:,:,0]*ones

    tens_yuv[:,:,0] = clahe.apply(tens_yuv[:,:,0])
    tens_yuv[:,:,0] = cv.equalizeHist(tens_yuv[:,:,0])
    tens_yuv[:,:,0] = clahe.apply(tens_yuv[:,:,0])

    tens_res = cv.cvtColor(tens_yuv, cv.COLOR_YUV2BGR)
    tens = to_tensor(tens_res, cspace='bgr')
    tens = _cvt_color(tens, cspace)

    return cv.fastNlMeansDenoisingColored(tens, None, 3, 3, 7, 21)


def _cvt_color(tens, cspace):
    if cspace == 'rgb':
        return to_rgb(tens)
    if cspace == 'bgr':
        return to_bgr(tens)
    if cspace == 'gray':
        return to_gray(tens)
    if cspace == 'hsv':
        return to_hsv(tens)
    if cspace == 'hls':
        return to_hls(tens)
    if cspace == 'lab':
        return to_lab(tens)


def _generate_shadow_coordinates(num_shadows, rectangular_roi, shadow_dimension):
    vertices_list=[]
    x1 = rectangular_roi[0]
    y1 = rectangular_roi[1]
    x2 = rectangular_roi[2]
    y2 = rectangular_roi[3]

    for index in range(num_shadows):
        vertex = []

        for dimensions in range(shadow_dimension): ## Dimensionality of the shadow polygon
            vertex.append((random.randint(x1, x2), random.randint(y1, y2)))

        vertices = np.array([vertex], dtype=np.int32) ## single shadow vertices 
        vertices_list.append(vertices)
    return vertices_list ## List of shadow vertices


def _shadow_process(tens, num_shadows, x1, y1, x2, y2, shadow_dimension) -> Tensor:
    tens = to_tensor(tens, enforce_tensor=True)
    tens = _hls(tens) ## Conversion to hls
    cspace = tens.cspace 

    mask = np.zeros_like(tens) 
    imshape = tens.shape

    # Get the list of shadow vertices
    vertices_list = _generate_shadow_coordinates(num_shadows, (x1,y1,x2,y2), shadow_dimension) 

    for vertices in vertices_list: 
        cv.fillPoly(mask, vertices, 255) ## adding all shadow polygons on empty mask, single 255 denotes only red channel

    # If red channel is hot, the Tensor's "Lightness" channel's brightness value is lowered 
    tens[:,:,1][mask[:,:,0]==255] = tens[:,:,1][mask[:,:,0]==255]*0.5   

    return to_tensor(tens, cspace=cspace)



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import cv2 as cv
import numpy as np
import gc

from .path import listdir, join


__all__ = [
    'get_classes_from_dir',
    'median',
    'train_val_split',
    'sort_dict',
    'get_opencv_version',
    'mean',
    'merge',
    'edges'
]


def median(arr, axis=None):
    return np.median(arr, axis=axis)

 
def mean(image, mask=None):
    return cv.mean(image, mask=mask)


def merge(tens):
    # if not isinstance(tens, (list, caer.Tensor)):
    #     raise ValueError('tens must be a list or caer.Tensor of (ideally) shape = 3)')

    return cv.merge(tens)

def get_classes_from_dir(DIR, verbose=0):
    if len(listdir(DIR, verbose=0)) == 0:
        raise ValueError('The specified directory does not seem to have any folders in it')
    else:
        import os 
        classes = [i for i in listdir(DIR, recursive=False, verbose=verbose) if os.path.isdir(join(DIR, i))]
        return classes


def _sep(data):
    
    x = [i[0] for i in data]
    y = [i[1] for i in data]

    return x, y


def train_val_split(X, y, val_ratio=.2):
    """
        Returns X_train, X_val, y_train, y_val
    """
    if len(X) != len(y):
        raise ValueError('X must be equal to y')
    
    data = [] 
    for i in range(len(X)):
        data.append([X[i], y[i]])

    split = int(len(X) - (len(X) * val_ratio)) - 1

    data_train = data[0:split]
    data_val = data[split:]

    X_train, y_train = (np.array(item) for item in _sep(data_train))
    X_val, y_val = (np.array(item) for item in _sep(data_val))

    del data
    gc.collect()

    return X_train, X_val, y_train, y_val


def sort_dict(unsorted_dict, descending=False):
    """ 
        Sorts a dictionary in ascending order (if descending = False) or descending order (if descending = True)
    """
    if not isinstance(descending, bool):
        raise ValueError('`descending` must be a boolean')

    return sorted(unsorted_dict.items(), key=lambda x:x[1], reverse=descending)


def get_opencv_version():
    return cv.__version__[0]


def edges(tens, threshold1=None, threshold2=None, use_median=True, sigma=None):
    if not isinstance(use_median, bool):
        raise ValueError('use_median must be a boolean')

    if not isinstance(threshold1, int) or not isinstance(threshold2, int):
        raise ValueError('Threshold values must be integers')

    if tens is None:
        raise ValueError('Image is of NoneType()')

    if not use_median and (threshold1 is None or threshold2 is None):
        raise ValueError('Specify valid threshold values')
    
    if use_median:
        if sigma is None:
            sigma = .3

        # computes the median of the single channel pixel intensities
        med = median(tens)

        # Canny edge detection using the computed mean
        low = int(max(0, (1.0-sigma) * med))
        up = int(min(255, (1.0+sigma) * med))
        canny_edges = cv.Canny(tens, low, up)
    
    else:
        canny_edges = cv.Canny(tens, threshold1, threshold2)

    return canny_edges



# This file is automatically generated during the generation of setup.py
# Copyright 2020-2021, Caer
author = 'Jason Dsouza'
author_email = 'jasmcaus@gmail.com'
version = '2.0.6'
full_version = '2.0.6'
git_version = '3e589e5456d7fe3efab234caf322e8ae880974bb'
release = True
contributors = ['Jason Dsouza <@jasmcaus>', '', 'Special Thanks to:', '@MalikMAlna', '@lucasace', '@GitHubDragonFly']
homepage = 'https://github.com/jasmcaus/caer'



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


from .extract_frames import (
    extract_frames,
    __all__ as __all_extract__
)

from .livestream import (
    LiveStream,
    __all__ as __all_live__ 
)

from .stream import (
    Stream,
    __all__ as __all_str__
)

from .gpufilestream import (
    GPUFileStream,
    __all__ as __all_gpu__ 
)

from .frames_and_fps import (
    count_frames,
    get_fps,
    __all__ as __all_ffps__
)


__all__ = __all_extract__ + __all_str__ + __all_live__ + __all_ffps__ + __all_gpu__


# Stop polluting the namespace
del __all_extract__
del __all_str__
del __all_ffps__
del __all_gpu__
del __all_live__


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


__all__ = [
    'FPS',
    'FRAME_COUNT',
    'FRAME_WIDTH',
    'FRAME_HEIGHT'
]


## Video
FRAME_WIDTH = 3
FRAME_HEIGHT = 4
FPS = 5
FRAME_COUNT = 7  


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

import numpy as np 
from .tensor import Tensor 

__all__ = [
    'from_numpy',
    'to_tensor'
]


def from_numpy(x, cspace, dtype=None, override_checks=False, enforce_tensor=False):
    r"""
        Convert a Numpy Array to a Caer Tensor. 
    
    .. warning::
        ``override_checks`` and ``enforce_tensor`` should not be used. If not handled property, a silly bug may pop up at an unexpected time.

    Args:
        x (ndarray): Array to convert.
        cspace (str): ``must`` be specified. Value should be either bgr/rgb/gray/hsv/hls/lab.
        dtype (numpy): (optional) Data Type 
        override_checks (bool): WARNING: Intended for internal Usage only => if not handled properly, a silly bug may pop up at an unexpected time.
        enforce_tensor (bool): WARNING: Intended for internal Usage only => if not handled properly, a silly bug may pop up at an unexpected time.
    
    Returns:
        ``caer.Tensor``
    """

    if cspace is None:
        if not override_checks:
            raise ValueError('The `cspace` attribute must be specified when converting from foreign arrays to a caer.Tensor. Supported values: bgr/rgb/gray/hsv/hls/lab.')
        
        # For certain functions, it is imperative that a caer.Tensor must be passed
        if enforce_tensor:
            raise ValueError('To use this function, a caer.Tensor must be passed')

    if isinstance(x, np.ndarray):
        return Tensor(x, cspace=cspace, dtype=dtype)
    
    else:
        raise TypeError('`x` is not a Numpy Array')


def to_tensor(x, cspace=None, dtype=None, override_checks=False, enforce_tensor=False):
    r"""
        Convert an array to a caer.Tensor.

        To convert a Numpy Array to a ``caer.Tensor``, specify the ``cspace`` attribute (either bgr/rgb/gray/hsv/hls/lab).
        If a ``caer.Tensor`` is passed, its attributes are preserved.

        If ``x`` is a ``caer.Tensor`` and ``cspace`` is specified, the colorspace of the Tensor will be updated accordingly. To prevent this, set ``cspace=None``. 

    .. warning::
        ``override_checks`` and ``enforce_tensor`` should not be used. If not handled property, a silly bug may pop up at an unexpected time.

    Args:
        x (Tensor, ndarray, PIL): Tensor/Array to convert.
        cspace (str): ``must`` be specified if you are converting a Numpy array to a ``caer.Tensor``. Value should be either bgr/rgb/gray/hsv/hls/lab.
        dtype (numpy): (optional) Data Type 
        override_checks (bool): WARNING: Intended for internal Usage only => if not handled properly, a silly bug may pop up at an unexpected time.
        enforce_tensor (bool): WARNING: Intended for internal Usage only => if not handled properly, a silly bug may pop up at an unexpected time.
    """
    if isinstance(x, Tensor):
        if cspace is None: # if unsure of the cspace (usually in cases when converting a foreign array to a caer.Tensor)
            return x 

        else:
            # Preserve
            x = Tensor(x, cspace=cspace, dtype=dtype)
            tens = _preserve_tensor_attrs(old=x)
            return tens 

    elif isinstance(x, np.ndarray):
        return from_numpy(x, cspace=cspace, override_checks=override_checks, enforce_tensor=enforce_tensor)

    # If PIL Image 
    elif 'PIL' in str(type(x)):
        if cspace is None:
            raise ValueError('The `cspace` attribute must be specified when converting from foreign arrays to a caer.Tensor. Supported values: bgr/rgb/gray/hsv/hls/lab.')

        x = np.array(x)
        return Tensor(x, cspace=cspace, dtype=dtype)

    else:
        raise TypeError(f'Cannot convert class {type(x)} to a caer.Tensor. Currently, only Numpy arrays and (to a limited extend) PIL images) are supported.')


def _preserve_tensor_attrs(old):
    r"""
        Copies Tensor attributes (like self.cspace) of `old` in `new`. Both must be caer.Tensors 
    
    Args:
        old (Tensor): caer Tensor
    
    Returns:
        caer Tensor
    """
    if not isinstance(old, Tensor):
        raise TypeError('`old` needs to be a caer.Tensor.')

    new = Tensor(old, cspace=None, dtype=old.dtype)
    new.cspace = old.cspace 
    return new 


# def _convert_to_tensor_and_rename_cspace(x, to) -> Tensor:
#     r"""
#     Args:
#         x (Tensor, ndarray): Image Tensor
#         to (str): rgb/bgr/gray/hsv/hls/lab
        
#     Returns:
#         Tensor
#     """
#     return to_tensor(x, cspace=to)


# # def _assign_mode_to_tensor(self, rgb, gray, mode=None):
# #     r"""
# #         Assign proper value of self.cspace 

# #         Idea:
# #             rgb = True ==> self.cspace = 'rgb'
# #             rgb = False ==> self.cspace = 'bgr'
# #             gray = True ==> self.cspace = 'gray'

# #         WARNING:
# #             Use `mode` explicitely ONLY if you are inside a function that converts to HSV, HLS or LAB.
# #     """
# #     if not isinstance(rgb, bool):
# #         raise TypeError('`rgb` needs to be boolean')

# #     if not isinstance(gray, bool):
# #         raise TypeError('`gray` needs to be boolean')
    
# #     if mode is not None and isinstance(mode, str):
# #         self.cspace = mode 
    


#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

import cv2 as cv

from ..coreten import Tensor, to_tensor
from ._constants import LUV2BGR, LUV2RGB
from ._bgr import bgr2gray, bgr2hls, bgr2hsv, bgr2lab, bgr2yuv

__all__ = [
    'luv2bgr',
    'luv2gray',
    'luv2hls',
    'luv2hsv',
    'luv2lab',
    'luv2rgb',
    'luv2yuv',
]


def _is_luv_image(tens):
    return len(tens.shape) == 3 and tens.shape[-1] == 3


def luv2bgr(tens) -> Tensor:
    r'''
        Converts an LUV Tensor to its BGR version.

    Args:
        tens (Tensor): Valid LUV Tensor

    Returns:
        BGR Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_luv_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts an LUV Tensor to its BGR counterpart'
        )

    im = cv.cvtColor(tens, LUV2BGR)
    return to_tensor(im, cspace='bgr')


def luv2rgb(tens) -> Tensor:
    r'''
        Converts a LUV Tensor to its RGB version.

    Args:
        tens (Tensor): Valid LUV Tensor

    Returns:
        RGB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_luv_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a LUV Tensor to its RGB counterpart'
        )

    im = cv.cvtColor(tens, LUV2RGB)
    return to_tensor(im, cspace='rgb')


def luv2gray(tens) -> Tensor:
    r'''
        Converts an LUV Tensor to its GRAY version.

    Args:
        tens (Tensor): Valid LUV Tensor

    Returns:
        GRAY Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_luv_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts an LUV Tensor to its GRAY counterpart'
        )

    im = luv2bgr(tens)
    im = bgr2gray(tens)
    return to_tensor(im, cspace='gray')


def luv2hls(tens) -> Tensor:
    r'''
        Converts an LUV Tensor to its HLS version.

    Args:
        tens (Tensor): Valid LUV Tensor

    Returns:
        HLS Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_luv_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts an LUV Tensor to its HLS counterpart'
        )

    im = luv2bgr(tens)
    im = bgr2hls(tens)
    return to_tensor(im, cspace='hls')


def luv2hsv(tens) -> Tensor:
    r'''
        Converts an LUV Tensor to its HSV version.

    Args:
        tens (Tensor): Valid LUV Tensor

    Returns:
        HSV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_luv_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts an LUV Tensor to its HSV counterpart'
        )

    im = luv2bgr(tens)
    im = bgr2hsv(tens)
    return to_tensor(im, cspace='hsv')


def luv2lab(tens) -> Tensor:
    r'''
        Converts an LUV Tensor to its LAB version.

    Args:
        tens (Tensor): Valid LUV Tensor

    Returns:
        LAB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_luv_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts an LUV Tensor to its LAB counterpart'
        )

    im = luv2bgr(tens)
    im = bgr2lab(tens)
    return to_tensor(im, cspace='lab')


def luv2yuv(tens) -> Tensor:
    r'''
        Converts an LUV Tensor to its YUV version.

    Args:
        tens (Tensor): Valid LUV Tensor

    Returns:
        YUV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_luv_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts an LUV Tensor to its YUV counterpart'
        )

    im = luv2bgr(tens)
    im = bgr2yuv(tens)
    return to_tensor(im, cspace='yuv')



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


#pylint:disable=pointless-string-statement

from .._internal import _check_mean_sub_values
from ..path import exists, list_images
from ..io import imread 
from ..coreten import Tensor
from ..annotations import Tuple, List

import numpy as np
import cv2 as cv

"""
    Important notes:
    Mean subtract must be computed ONLY on the training set and then later applied on the validation/test set
"""

__all__ = [
    'MeanProcess',
    'compute_mean',
    'compute_mean_from_dir',
    'subtract_mean',
]

class MeanProcess:
    def __init__(self, mean_sub_values, channels) -> None:
        # mean_sub_values is a tuple
        flag = _check_mean_sub_values(mean_sub_values, channels)
        if flag:
            if channels == 3:
                self.rMean = mean_sub_values[0]
                self.gMean = mean_sub_values[1]
                self.bMean = mean_sub_values[2]
            else:
                self.bgrMean = mean_sub_values[0]

    def mean_preprocess(self, image, channels) -> Tensor:
        """
            Mean Subtraction performed per channel
            Mean must be calculated ONLY on the training set
        """

        if channels == 3:
            b, g, r = cv.split(image.astype('float32'))[:3]

            # Subtracting the mean
            r -= self.rMean
            g -= self.gMean
            b -= self.bMean

            # Merging 
            return cv.merge([b,g,r])
        
        if channels == 1:
            image -= self.bgrMean
            return image
            

def compute_mean_from_dir(DIR, channels, per_channel_subtraction=True, recursive=True) -> Tuple:
    """
        Computes mean per channel
        Mean must be computed ONLY on the train set
    """
    if not exists(DIR):
        raise ValueError('The specified directory does not exist')
    
    image_list = list_images(DIR, recursive=recursive, use_fullpath=True, verbose=0)

    if len(image_list) == 0:
        raise ValueError(f'No images found at {DIR}')

    if channels == 3:
        rMean, gMean, bMean = 0, 0, 0
    if channels == 1:
        bgrMean = 0

    count = 0

    for tens_filepath in image_list:
        count += 1
        tens = imread(tens_filepath, rgb=True)

        if channels == 3:
            b, g, r = cv.mean(tens.astype('float32'))[:3]
            rMean += r
            bMean += b
            gMean += g

        if channels == 1:
            bgrMean += np.mean(tens.astype('float32'))

    # Computing average mean
    if channels == 3:
        rMean /= count
        bMean /= count 
        gMean /= count

        if per_channel_subtraction:
            return rMean, bMean, gMean
        else:
            mean_of_means = (rMean + bMean + gMean) / 3
            return mean_of_means, mean_of_means, mean_of_means

    if channels == 1:
        bgrMean /= count
        return tuple([bgrMean])


def compute_mean(data, channels, per_channel_subtraction=True) -> Tuple:
    """
        Computes mean per channel over the train set and returns a tuple of dimensions=channels
        Train should not be normalized
    """
    if len(data) == 0:
        raise ValueError('Dataset is empty')
    
    if not isinstance(data, (list, Tensor)):
        raise ValueError('Dataset must be a list of size=number of images and shape=image shape')

    if channels == 3:
        rMean, gMean, bMean = 0,0,0
    if channels == 1:
        bgrMean = 0

    count = 0

    for tens in data:
        count += 1
        if channels == 3:
            b, g, r = cv.mean(tens.astype('float32'))[:3]
            rMean += r
            gMean += g
            bMean += b
        if channels == 1:
            bgrMean += np.mean(tens.astype('float32'))

    # Computing average mean
    if channels == 3:
        rMean /= count
        bMean /= count 
        gMean /= count
        if per_channel_subtraction:
            return rMean, bMean, gMean
        else:
            mean_of_means = (rMean + bMean + gMean) / 3
            return mean_of_means, mean_of_means, mean_of_means

    if channels == 1:
        bgrMean /= count
        return tuple([bgrMean])


def subtract_mean(data, channels, mean_sub_values) -> List[str]:
    """
        Per channel subtraction values computed from compute_mean() or compute_mean_from_dir()
        Subtracts mean from the validation set
    """

    mean_process = MeanProcess(mean_sub_values, channels)

    if len(data) == 0:
        raise ValueError('Dataset is empty')
    
    if not isinstance(data, (list, Tensor)):
        raise ValueError('Dataset must be a list of size = number of images and shape = image shape')

    data = [mean_process.mean_preprocess(tens, channels) for tens in data]
    
    return data


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


from .stream import Stream
from ..path import exists

__all__ = [
    'count_frames',
    'get_fps'
]


def count_frames(video_path):
    r"""
        Returns the number of frames in a video at ``video_path``.

    Args:
        video_path (str): Video Filepath
    
    Returns:
        Frame count 

    """
    if video_path is None:
        raise ValueError('Specify a valid video path')

    stream = Stream(video_path)
    frame_count = stream.count_frames()
    stream.release()
    return frame_count


def get_fps(video_path=None):
    r"""
        Returns the FPS in a video at ``video_path``.

    Args:
        video_path (str): Video Filepath
    
    Returns:
        FPS value. 

    """
    if video_path is None and not exists(video_path):
        raise ValueError('Specify a valid video path')

    stream = Stream(video_path)
    fps_count = stream.get_fps()
    stream.release()
    
    return fps_count


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


# We don't want to modify the root dir in any way 
# This gets disrupted if this file is called directly
# Checks are in place:

if __name__ != '__main__':
    from .path import dirname

    __curr__ = dirname(__file__).replace("\\", "/")


#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import cv2 as cv

from ..coreten import Tensor, to_tensor
from ._constants import BGR2RGB, BGR2GRAY, BGR2HSV, BGR2LAB, BGR2HLS, BGR2YUV, BGR2LUV

__all__ = [
    'bgr2gray',
    'bgr2hsv',
    'bgr2lab',
    'bgr2rgb',
    'bgr2hls',
    'bgr2yuv',
    'bgr2luv',
]


def _is_bgr_image(tens):
    # tens = to_tensor(tens)
    # return tens.is_bgr()
    return len(tens.shape) == 3 and tens.shape[-1] == 3


def bgr2rgb(tens) -> Tensor:
    r'''
        Converts a BGR Tensor to its RGB version.

    Args:
        tens (Tensor): Valid BGR Tensor

    Returns:
        RGB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_bgr_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a BGR Tensor to its RGB counterpart'
        )

    im = cv.cvtColor(tens, BGR2RGB)
    return to_tensor(im, cspace='rgb')


def bgr2gray(tens) -> Tensor:
    r'''
        Converts a BGR Tensor to its Grayscale version.

    Args:
        tens (Tensor): Valid BGR Tensor

    Returns:
        Grayscale Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_bgr_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a BGR Tensor to its Grayscale counterpart'
        )

    tens = to_tensor(tens)
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value

    im = cv.cvtColor(tens, BGR2GRAY)
    return to_tensor(im, cspace='gray')


def bgr2hsv(tens) -> Tensor:
    r'''
        Converts a BGR Tensor to its HSV version.

    Args:
        tens (Tensor): Valid BGR Tensor

    Returns:
        HSV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_bgr_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a BGR Tensor to its HSV counterpart'
        )

    im = cv.cvtColor(tens, BGR2HSV)
    return to_tensor(im, cspace='hsv')


def bgr2lab(tens) -> Tensor:
    r'''
        Converts a BGR Tensor to its LAB version.

    Args:
        tens (Tensor): Valid BGR Tensor

    Returns:
        LAB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_bgr_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a BGR Tensor to its LAB counterpart'
        )

    im = cv.cvtColor(tens, BGR2LAB)
    return to_tensor(im, cspace='lab')


def bgr2hls(tens) -> Tensor:
    r'''
        Converts a BGR Tensor to its HLS version.

    Args:
        tens (Tensor): Valid BGR Tensor

    Returns:
        HLS Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_bgr_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a BGR Tensor to its HLS counterpart'
        )

    im = cv.cvtColor(tens, BGR2HLS)
    return to_tensor(im, cspace='hls')


def bgr2yuv(tens) -> Tensor:
    r'''
        Converts a BGR Tensor to its YUV version.

    Args:
        tens (Tensor): Valid BGR Tensor

    Returns:
        YUV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_bgr_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a BGR Tensor to its YUV counterpart'
        )

    im = cv.cvtColor(tens, BGR2YUV)
    return to_tensor(im, cspace='yuv')


def bgr2luv(tens) -> Tensor:
    r'''
        Converts a BGR Tensor to its LUV version.

    Args:
        tens (Tensor): Valid BGR Tensor

    Returns:
        YUV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_bgr_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a BGR Tensor to its LUV counterpart'
        )

    im = cv.cvtColor(tens, BGR2LUV)
    return to_tensor(im, cspace='luv')



r""" \
Caer
=====
Caer is a *lightweight, scalable* Computer Vision library for high-performance AI research. It simplifies your approach
towards Computer Vision by abstracting away unnecessary boilerplate code giving you the flexibility to quickly prototype deep
learning models or research ideas.

Our design philosophy makes Caer ideal for students, researchers, hobbyists and even experts in the fields of Deep Learning
and Computer Vision.

Documentation: https://caer.readthedocs.io

Available subpackages
---------------------
augment
    Powerful image augmentation operations.
color
    Colorspace manipulation
data
    Standard high-quality test images
path
    OS-specific Path Manipulations
preprocessing
    Image preprocessing utilities
video
    Video processing utilities


Utilities
---------
__version__
    Caer version string

__author__
    Author of Caer

__contributors__
    List of all contributors to the project

__homepage__
    Web URL of the Caer documentation

"""

#pylint:disable=undefined-all-variable, redefined-builtin

from ._meta import (
    version,
    release,
    author,
    author_email,
    contributors,
    homepage
)

r"""
Root Package Info
"""

__version__ = version
__release__ = release
__author__ = author
__author_email__ = author_email
__contributors__ = contributors
__license__ = 'MIT License'
__copyright__ = r"""
Copyright (c) 2020-2021 Jason Dsouza <jasmcaus>
All Rights Reserved.
"""
__homepage__ = homepage


def license():
    return __license__

def copyright():
    return __copyright__


# IO-stuff
from .io import (
    imread,
    imsave,
    resize,
    smart_resize,
    __all__ as __all_io__
)

# Preprocessing
# Kept for backward compatibility (with my online OpenCV course published on FreeCodeCamp.org)
from .preprocess import (
    preprocess_from_dir,
    sep_train,
    normalize,
    __all__ as __all_preprocess__
)

# General utilities
from .core import (
    train_val_split,
    get_classes_from_dir,
    sort_dict,
    get_opencv_version,
    __all__ as __all_core__
)

# Color Spaces
from .color import (
    to_bgr,
    to_rgb,
    to_gray,
    to_hsv,
    to_hls,
    to_lab,
    to_yuv,
    to_luv,

    bgr2rgb,
    bgr2gray,
    bgr2hsv,
    bgr2hls,
    bgr2lab,
    bgr2yuv,
    bgr2luv,

    rgb2bgr,
    rgb2gray,
    rgb2hsv,
    rgb2hls,
    rgb2lab,
    rgb2yuv,
    rgb2luv,

    gray2rgb,
    gray2bgr,
    gray2hsv,
    gray2hls,
    gray2lab,
    gray2yuv,
    gray2luv,

    hsv2rgb,
    hsv2bgr,
    hsv2gray,
    hsv2hls,
    hsv2lab,
    hsv2yuv,
    hsv2luv,

    hls2rgb,
    hls2bgr,
    hls2gray,
    hls2hsv,
    hls2lab,
    hls2yuv,
    hls2luv,

    lab2rgb,
    lab2bgr,
    lab2gray,
    lab2hsv,
    lab2hls,
    lab2yuv,
    lab2luv,

    yuv2rgb,
    yuv2bgr,
    yuv2gray,
    yuv2hsv,
    yuv2hls,
    yuv2lab,
    yuv2luv,

    luv2rgb,
    luv2bgr,
    luv2gray,
    luv2hsv,
    luv2hls,
    luv2lab,
    luv2yuv,

    __all__ as __all_color__
)


# Bringing in configuration variables from globals.py
# from .globals import *

# Tensor-stuff
from .coreten import *


def get_caer_version():
    return __version__


def get_caer_functions():
    return __all__


def get_caer_methods():
    return __all__


# __all__ configs
from .globals import __all__ as __all_globals__

from .transforms import __all__ as __all_transforms__
from .data import __all__ as __all_data__
from .path import __all__ as __all_path__
from .preprocessing import __all__ as __all_preprocessing__
from .video import __all__ as __all_video__


__all__ = __all_globals__ + __all_core__ + __all_preprocess__

__all__ += __all_transforms__
__all__ += __all_color__
__all__ += __all_data__
__all__ += __all_io__
__all__ += __all_path__
__all__ += __all_preprocessing__
__all__ += __all_video__


# Stop polluting the namespace

## Remove root package info
del author
del version
del release
del contributors

## Remove everything else
del __all_globals__
del __all_core__
del __all_preprocess__

del __all_transforms__
del __all_color__
del __all_data__
del __all_preprocessing__
del __all_video__



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import math
import numpy as np

class __PrinterOptions(object):
    precision = 4
    threshold = 1000
    edgeitems = 3
    linewidth = 80
    sci_mode = None
    prefix = 'tensor('


PRINT_OPTS = __PrinterOptions()


class _Formatter(object):
    def __init__(self, tensor):
        self.floating_dtype = 'float' in str(repr(tensor.dtype))
        self.int_mode = True
        self.sci_mode = False
        self.max_width = 1

        tensor_view = tensor.reshape(-1)

        if not self.floating_dtype:
            for value in tensor_view:
                value_str = '{}'.format(value)
                # d = max(self.max_width, len(value_str))
                d = max(2, len(value_str))
                self.max_width = d

        else:
            # FLOATING POINT

            for value in tensor_view:
                if value != np.ceil(value):
                    self.int_mode = False
                    break

            if self.int_mode:
                for value in tensor_view:
                    value_str = ('{:.0f}').format(value)
                    self.max_width = max(self.max_width, len(value_str) + 1)


        if PRINT_OPTS.sci_mode is not None:
            self.sci_mode = PRINT_OPTS.sci_mode


    def width(self):
        return self.max_width


    def format(self, value):
        if self.floating_dtype:
            if self.sci_mode:
                ret = ('{{:{}.{}e}}').format(self.max_width, PRINT_OPTS.precision).format(value)

            elif self.int_mode:
                ret = '{:.0f}'.format(value)
                if not (math.isinf(value) or math.isnan(value)):
                    ret += '.'

            else:
                ret = ('{{:.{}f}}').format(PRINT_OPTS.precision).format(value)

        else:
            ret = '{}'.format(value)

        # return (self.max_width - len(ret)) * ' ' + ret
        return (2 - len(ret)) * ' ' + ret

def _scalar_str(self, formatter):
    # Usually, we must never come here. 
    # This is only for when the 'coreten' library is built.
    # Changes may be made to this.
    return formatter.format(self.item())


def _vector_str(self, indent, summarize, formatter):
    # length includes spaces and comma between elements
    element_length = formatter.width() + 2

    elements_per_line = max(1, int(math.floor((PRINT_OPTS.linewidth - indent) / (element_length))))
    char_per_line = element_length * elements_per_line

    def _val_formatter(val, formatter=formatter):
        return formatter.format(val)

    # Preventing the entire tensor from being displayed to the terminal. 
    # We (figuratively) "prune" the tensor for output
    if summarize and self.size_dim(0) > 2 * PRINT_OPTS.edgeitems:
        data = ([_val_formatter(val) for val in self[:PRINT_OPTS.edgeitems].tolist()] +
                [' ...'] +
                [_val_formatter(val) for val in self[-PRINT_OPTS.edgeitems:].tolist()])
    else:
        data = [_val_formatter(val) for val in self.tolist()]

    data_lines = [data[i:i + elements_per_line] for i in range(0, len(data), elements_per_line)]
    lines = [', '.join(line) for line in data_lines]

    return '[' + (',' + '\n' + ' ' * (indent + 1)).join(lines) + ']'


def _tensor_str_with_formatter(self, indent, summarize, formatter):
    dim = self.dim()
    # dim = self.ndim()

    # if dim == 0:
    #     return _scalar_str(self, formatter)

    if dim == 1:
        return _vector_str(self, indent, summarize, formatter)

    # Preventing the entire tensor from being displayed to the terminal. 
    # We (figuratively) "prune" the tensor for output
    if summarize and self.size_dim(0) > 2 * PRINT_OPTS.edgeitems:
        slices = ([_tensor_str_with_formatter(self[i], indent + 1, summarize, formatter)
                   for i in range(0, PRINT_OPTS.edgeitems)] +
                  ['...'] +
                  [_tensor_str_with_formatter(self[i], indent + 1, summarize, formatter)
                   for i in range(len(self) - PRINT_OPTS.edgeitems, len(self))])

    # If tensor is small enough to display to terminal
    else:
        slices = [_tensor_str_with_formatter(self[i], indent + 1, summarize, formatter)
                  for i in range(0, self.size_dim(0))]

    tensor_str = (',' + '\n' * (dim - 1) + ' ' * (indent + 1)).join(slices)
    return '[' + tensor_str + ']'


def _tensor_str(self, indent):
    if self.numel() == 0:
        return '[]'
    summarize = self.numel() > PRINT_OPTS.threshold
    # summarize = self.size_dim > PRINT_OPTS.threshold

    # if self.dtype is torch.float16 or self.dtype is torch.bfloat16:
    #     self = self.float()


    formatter = _Formatter(get_summarized_data(self) if summarize else self)

    x = _tensor_str_with_formatter(self, indent, summarize, formatter)

    return x


def _add_suffixes(tensor_str, suffixes, indent, force_newline):
    tensor_strs = [tensor_str]
    last_line_len = len(tensor_str) - tensor_str.rfind('\n') + 1

    for suffix in suffixes:
        suffix_len = len(suffix)

        if force_newline or last_line_len + suffix_len + 2 > PRINT_OPTS.linewidth:
            tensor_strs.append(',\n' + ' ' * indent + suffix)
            last_line_len = indent + suffix_len
            force_newline = False
        else:
            tensor_strs.append(', ' + suffix)
            last_line_len += suffix_len + 2

    tensor_strs.append(')')

    return ''.join(tensor_strs)


def get_summarized_data(self):
    
    dim = self.dim()
    
    if dim == 0:
        return self

    if dim == 1:
        if self.size_dim(0) > 2 * PRINT_OPTS.edgeitems:
            return np.concatenate((self[:PRINT_OPTS.edgeitems], self[-PRINT_OPTS.edgeitems:]))
        else:
            return self

    if self.size_dim(0) > 2 * PRINT_OPTS.edgeitems:
        start = [self[i] for i in range(0, PRINT_OPTS.edgeitems)]
        end = ([self[i]
               for i in range(len(self) - PRINT_OPTS.edgeitems, len(self))])
        return np.stack([get_summarized_data(x) for x in (start + end)])
    else:
        return np.stack([get_summarized_data(x) for x in self])


def _str_intern(self):
    
    prefix = PRINT_OPTS.prefix
    indent = len(prefix)
    suffixes = []


    if self.numel() == 0:
        # Explicitly print the shape if it is not (0,), to match NumPy behavior
        if self.dim() != 1:
            suffixes.append('size=' + str(tuple(self.shape)))

        tensor_str = '[]'

    else:
        suffixes.append('dtype=' + str(self.dtype))

        tensor_str = _tensor_str(self, indent)

    return _add_suffixes(prefix + tensor_str, suffixes, indent, force_newline=False)


def _str(self):
    return _str_intern(self)



""" cythonize
    Cythonize pyx files into C files as needed.
    Usage: cythonize [root_dir]
    Default [root_dir] is '../caer'.
    
    Simple script to invoke Cython on all .pyx files; while waiting for a proper build system. 

    For now, this script should be run by developers when changing Cython files
    only, and the resulting C files checked in, so that end-users (and Python-only
    developers) do not get the Cython dependencies.
    
    Inspired by the script originally written by Dag Sverre Seljebotn, and copied here from:
    https://raw.github.com/dagss/private-scipy-refactor/cythonize/cythonize.py
    Note: this script does not check any of the dependent C libraries; it only
    operates on the Cython .pyx files.

    The above script compares pyx files to see if they have been changed relative to their corresponding C files by comparing hashes stored in a database file. It calls `cython [file.pyx]` which merely converts the .pyx files to .c files. This cannot be imported into a Python file (an extension needs to be built for that). 

    This current script uses a command which converts the .pyx --> .c which then builds the required extensions (.pyd on Windows).

    To manually build the required extensions from .c files, use the following code:
    
    # from distutils.core import setup, Extension

    # module1 = Extension('demo',
    #                     include_dirs = ['/usr/local/include'],
    #                     libraries = ['tcl83'],
    #                     library_dirs = ['/usr/local/lib'],
    #                     sources = ['demo.c'])

    # setup (
        # # name = 'PackageName',
        # # version = '1.0',
        # # description = 'This is a demo package',
        # # author = 'Martin v. Loewis',
        # # author_email = 'martin@v.loewis.de',
        # # url = 'https://docs.python.org/extending/building',
        # # long_description = '''
        # #       This is really just a demo package.
        # #       ''',
    #     ext_modules = [module1])
"""


import os 
import sys 
import datetime
import subprocess 

#pylint:disable=redefined-builtin, bare-except

ROOT_DIR = '../caer'

CYTHON_SOURCES = []

# WindowsError is not defined on unix systems
try:
    WindowsError
except NameError:
    WindowsError = None

# This is a bit hackish. We are auto-generating a python file that will allow for Cython to build the required extensions.
SETUP_TEXT =\
"""
# DO NOT EDIT!
# THIS FILE WAS AUTOMATICALLY GENERATED AT %(time)s DURING THE CYTHON BUILD CYCLE OF ALL .PYX FILES
# Copyright 2020-2021, Caer

from setuptools import setup
from Cython.Build import cythonize

setup(
    ext_modules=cythonize(%(source)s),
    zip_safe=False,
)
"""


#
# Rules
#
def find_files(path, *ext):
    for root, _, files in os.walk(path):
        if 'caer' in root and not '.git' in root:
            for file in files:
                if file.endswith(ext):
                    base = os.path.join(root,file).replace('\\', '/')
                    abspath = os.path.abspath(base)
                    CYTHON_SOURCES.append(abspath)


def normpath(path):
    path = path.replace(os.sep, '/')
    if path.startswith('./'):
        path = path[2:]
    elif path.startswith('../'):
        path = path[3:]
    return path


def process_pyx():
    flags = ['-3', '--inplace']
    build_cython = True

    try:
        from Cython.Compiler.Version import version as cython_version
    except ImportError:
        raise OSError('[ERROR] Cython needs to be installed')

    # Cython 0.29.21 is required for Python 3.9 and there are other fixes in the 0.29 series that are needed for earlier Python versions.
    # Note: keep in sync with that in pyproject.toml
    pyproject_toml = os.path.join(os.path.dirname(__file__), '..', 'pyproject.toml')
    if not os.path.exists(pyproject_toml):
        raise RuntimeError('[ERROR] pyproject.toml was not found. Ensure it was not deleted')

    # Try to find the minimum version from pyproject.toml
    # required_cython_version = '0.29.21'
    with open(pyproject_toml) as f:
        for line in f:
            if 'cython' not in line.lower():
                continue
            _, line = line.split('=')
            required_cython_version, _ = line.split("'")
            break
        else:
            raise RuntimeError('[ERROR] An appropriate version for Cython could not be found from pyproject.toml')

    if cython_version < required_cython_version:
        raise RuntimeError(f'[ERROR] Building Caer requires Cython >= {required_cython_version}')

    # Populating CYTHON_SOURCES
    find_files(ROOT_DIR, '.pyx')

    if len(CYTHON_SOURCES) == 0:
        sys.stderr.write('[ERROR] No Cython files found matching the required extensions. Cython build escaping\n')
        build_cython = False

    if build_cython:
        # Writing to build_cython.py (temp)
        a = open('tools/build_cython.py', 'w')
        try:
            a.write(SETUP_TEXT % {'time': datetime.date.today().strftime("%B %d, %Y") ,
                                  'source': CYTHON_SOURCES } )
        finally:
            a.close()


        # Can only concatenate lists
        # build_cython.py is an auto-generated Python script by cythonize.py
        # Do not run cythonize.py directly unless the .c files need to be re-compiled
        try:
            r = subprocess.check_call(['python' + 'build_cython.py' + 'build_ext'] + flags)

            if r != 0:
                p = subprocess.check_call([sys.executable, '-m', 'build_cython.py'] + flags)

                if p != 0:
                    raise Exception()

        except:
            sys.stderr.write('[ERROR] Building cython failed')



def main():
    # try:
    #     ROOT_DIR = sys.argv[1]
    # except IndexError:
    #     ROOT_DIR = ROOT_DIR
    process_pyx()


if __name__ == '__main__':
    main()




# """ cythonize
# Cythonize pyx files into C files as needed.
# Usage: cythonize [root_dir]
# Default [root_dir] is 'caer'.
# Checks pyx files to see if they have been changed relative to their
# corresponding C files.  If they have, then runs cython on these files to
# recreate the C files.
# The script thinks that the pyx files have changed relative to the C files
# by comparing hashes stored in a database file.
# Simple script to invoke Cython (and Tempita) on all .pyx (.pyx.in)
# files; while waiting for a proper build system. Uses file hashes to
# figure out if rebuild is needed.
# For now, this script should be run by developers when changing Cython files
# only, and the resulting C files checked in, so that end-users (and Python-only
# developers) do not get the Cython/Tempita dependencies.
# Originally written by Dag Sverre Seljebotn, and copied here from:
# https://raw.github.com/dagss/private-scipy-refactor/cythonize/cythonize.py
# Note: this script does not check any of the dependent C libraries; it only
# operates on the Cython .pyx files.
# """
# #pylint:disable=redefined-builtin

# import os
# import re
# import sys
# import hashlib
# import subprocess

# HASH_FILE = 'cythonize.dat'
# ROOT_DIR = 'caer'
# VENDOR = 'caer'

# CYTHON_SOURCES = []

# # WindowsError is not defined on unix systems
# try:
#     WindowsError
# except NameError:
#     WindowsError = None

# #
# # Rules
# #
# def find_files(*ext):
#     for root, _, files in os.walk('..'):
#         if 'caer' in root and not '.git' in root:
#             for file in files:
#                 if file.endswith(ext):
#                     fi = root + '\\' + file
#                     CYTHON_SOURCES.append(fi.replace('\\', '/')[3:])

# def process_pyx():
#     flags = ['-3', '--inplace']

#     try:
#         # try the cython in the installed python first (somewhat related to scipy/scipy#2397)
#         from Cython.Compiler.Version import version as cython_version
#     except ImportError:
#         raise OSError('Cython needs to be installed')

#     # Cython 0.29.21 is required for Python 3.9 and there are
#     # other fixes in the 0.29 series that are needed even for earlier
#     # Python versions.
#     # Note: keep in sync with that in buildproject.toml
#     required_cython_version = '0.29.21'

#     if cython_version < required_cython_version:
#         raise RuntimeError(f'Building Caer requires Cython >= {required_cython_version}')
#     # subprocess.check_call(
#     #     [sys.executable, '-m', 'cython'] + flags + ["-o", tofile, fromfile])

#     find_files('.py')
#     # Can only concatenate lists
#     subprocess.check_call(
#         [sys.executable, '-m', 'cythonize'] + flags + CYTHON_SOURCES)



# # def process_tempita_pyx(fromfile, tofile):
# #     import npy_tempita as tempita

# #     assert fromfile.endswith('.pyx.in')
# #     with open(fromfile, "r") as f:
# #         tmpl = f.read()
# #     pyxcontent = tempita.sub(tmpl)
# #     pyxfile = fromfile[:-len('.pyx.in')] + '.pyx'
# #     with open(pyxfile, "w") as f:
# #         f.write(pyxcontent)
# #     process_pyx(pyxfile, tofile)


# # def process_tempita_pyd(fromfile, tofile):
# #     import npy_tempita as tempita

# #     assert fromfile.endswith('.pxd.in')
# #     assert tofile.endswith('.pxd')
# #     with open(fromfile, "r") as f:
# #         tmpl = f.read()
# #     pyxcontent = tempita.sub(tmpl)
# #     with open(tofile, "w") as f:
# #         f.write(pyxcontent)

# # def process_tempita_pxi(fromfile, tofile):
# #     import npy_tempita as tempita

# #     assert fromfile.endswith('.pxi.in')
# #     assert tofile.endswith('.pxi')
# #     with open(fromfile, "r") as f:
# #         tmpl = f.read()
# #     pyxcontent = tempita.sub(tmpl)
# #     with open(tofile, "w") as f:
# #         f.write(pyxcontent)

# # def process_tempita_pxd(fromfile, tofile):
# #     import npy_tempita as tempita

# #     assert fromfile.endswith('.pxd.in')
# #     assert tofile.endswith('.pxd')
# #     with open(fromfile, "r") as f:
# #         tmpl = f.read()
# #     pyxcontent = tempita.sub(tmpl)
# #     with open(tofile, "w") as f:
# #         f.write(pyxcontent)

# # rules = {
# #     # fromext : function, toext
# #     '.pyx' : (process_pyx, '.c'),
# #     '.pyx.in' : (process_tempita_pyx, '.c'),
# #     '.pxi.in' : (process_tempita_pxi, '.pxi'),
# #     '.pxd.in' : (process_tempita_pxd, '.pxd'),
# #     '.pyd.in' : (process_tempita_pyd, '.pyd'),
# #     }

# rules = {
#     # fromext : function, toext
#     '.pyx' : (process_pyx, '.c')
# }

# #
# # Hash db
# #
# def load_hashes(filename):
#     # Return { filename : (sha1 of input, sha1 of output) }
#     if os.path.isfile(filename):
#         hashes = {}
#         with open(filename, 'r') as f:
#             for line in f:
#                 filename, inhash, outhash = line.split()
#                 hashes[filename] = (inhash, outhash)
#     else:
#         hashes = {}
#     return hashes


# def save_hashes(hash_db, filename):
#     with open(filename, 'w') as f:
#         for key, value in sorted(hash_db.items()):
#             f.write("%s %s %s\n" % (key, value[0], value[1]))


# def sha1_of_file(filename):
#     h = hashlib.sha1()
#     with open(filename, "rb") as f:
#         h.update(f.read())
#     return h.hexdigest()


# #
# # Main program
# #

# def normpath(path):
#     path = path.replace(os.sep, '/')
#     if path.startswith('./'):
#         path = path[2:]
#     return path


# def get_hash(frompath, topath):
#     from_hash = sha1_of_file(frompath)
#     to_hash = sha1_of_file(topath) if os.path.exists(topath) else None
#     return (from_hash, to_hash)


# def process(path, fromfile, tofile, processor_function, hash_db):
#     fullfrompath = os.path.join(path, fromfile)
#     fulltopath = os.path.join(path, tofile)
#     current_hash = get_hash(fullfrompath, fulltopath)
#     if current_hash == hash_db.get(normpath(fullfrompath), None):
#         print(f'{fullfrompath} has not changed')
#         return

#     orig_cwd = os.getcwd()
#     try:
#         os.chdir(path)
#         print(f'Processing {fullfrompath}')
#         processor_function(fromfile, tofile)
#     finally:
#         os.chdir(orig_cwd)
#     # changed target file, recompute hash
#     current_hash = get_hash(fullfrompath, fulltopath)
#     # store hash in db
#     hash_db[normpath(fullfrompath)] = current_hash


# def find_process_files(root_dir):
#     hash_db = load_hashes(HASH_FILE)
#     files  = [x for x in os.listdir(root_dir) if not os.path.isdir(x)]
#     # # .pxi or .pxi.in files are most likely dependencies for
#     # # .pyx files, so we need to process them first
#     # files.sort(key=lambda name: (name.endswith('.pxi') or
#     #                              name.endswith('.pxi.in') or
#     #                              name.endswith('.pxd.in')),
#     #            reverse=True)

#     for filename in files:
#         # in_file = os.path.join(root_dir, filename + ".in")
#         for fromext, value in rules.items():
#             if filename.endswith(fromext):
#                 if not value:
#                     break
#                 function, toext = value
#                 if toext == '.c':
#                     with open(os.path.join(root_dir, filename), 'rb') as f:
#                         data = f.read()
#                         m = re.search(br"^\s*#\s*distutils:\s*language\s*=\s*c\+\+\s*$", data, re.I|re.M)
#                         if m:
#                             toext = ".cxx"
#                 fromfile = filename
#                 tofile = filename[:-len(fromext)] + toext
#                 process(root_dir, fromfile, tofile, function, hash_db)
#                 save_hashes(hash_db, HASH_FILE)
#                 break

# def main():
#     try:
#         root_dir = sys.argv[1]
#     except IndexError:
#         root_dir = ROOT_DIR
#     find_process_files(root_dir)


# if __name__ == '__main__':
#     main()


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


__all__ = [
    'CROP_CENTRE',
    'CROP_TOP',
    'CROP_LEFT',
    'CROP_RIGHT',
    'CROP_BOTTOM',
    'INTER_NEAREST',
    'INTER_LINEAR',
    'INTER_AREA',
    'INTER_CUBIC',
]


# Configuration Variables used in Caer
CROP_CENTRE = 1
CROP_TOP = 2
CROP_LEFT = 3
CROP_RIGHT = 4
CROP_BOTTOM = 5

#  OpenCV 

## Interpolations
INTER_NEAREST = 0  # cv.INTER_NEAREST
INTER_LINEAR = 1  # cv.INTER_LINEAR
INTER_CUBIC = 2  # cv.INTER_CUBIC
INTER_AREA = 3  # cv.INTER_AREA



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

#pylint:disable=unused-argument,unused-variable,eval-used

import numpy as np 
import cv2 as cv 
import random
import math 

from ..coreten import Tensor, is_tensor

from .functional import (
    is_numeric, 
    is_numeric_list_or_tuple, 
    is_tuple, 
    _rgb,
    _snow_process,
    _generate_random_lines,
    _rain_process,
    _generate_random_blur_coordinates,
    _add_blur,
    _gravel_process,
    _add_sun_flare_line,
    _add_sun_process,
    _apply_motion_blur,
    _autumn_process,
    _shadow_process
)

__all__ = [
    'sim_snow',
    'sim_rain',
    'sim_fog',
    'sim_gravel',
    'sim_sun_flare',
    'sim_motion_blur',
    'sim_autumn',
    'sim_shadow'
]

def sim_snow(tens, snow_coeff=-1) -> Tensor:
    r"""
        Simulate snowy conditions on an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        snow_coeff (int): Coefficient value.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_snow(tens, snow_coeff=-1)
        >> filtered
        (427, 640, 3)

    """
    if snow_coeff != -1:
        if snow_coeff < 0.0 or snow_coeff > 1.0:
            raise ValueError('Snow coeff must only be between 0 and 1')
    else:
        snow_coeff = random.uniform(0, 1)

    snow_coeff *= 255/2
    snow_coeff += 255/3

    return _snow_process(tens, snow_coeff)
    


# Rain_type = 'drizzle', 'heavy', 'torrential'
def sim_rain(tens, slant=-1, drop_length=20, drop_width=1, drop_color=(200,200,200), rain_type='None') -> Tensor: ## (200,200,200) is a shade of gray
    r"""
        Simulate rainy conditions on an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        slant (int): Slant value.
        drop_length (int): Length of the raindrop.
        drop_width (int): Width of the raindrop.
        drop_color (tuple): Color of the raindrop.
        rain_type (str): Type of rain. Can be either 'drizzle', 'heavy' or 'torrential'.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_rain(tens)
        >> filtered
        (427, 640, 3)

    """
    
    slant_extreme = slant
    if not(is_numeric(slant_extreme) and (slant_extreme >=-20 and slant_extreme <= 20) or slant_extreme==-1):
        raise ValueError('Numeric value must be between -20 and 20')

    if not(is_numeric(drop_width) and drop_width>=1 and drop_width<=5):
        raise ValueError('Width must be between 1 and 5')

    if not(is_numeric(drop_length) and drop_length>=0 and drop_length<=100):
        raise ValueError('Length must be between 0 and 100')

    imshape = tens.shape
    if slant_extreme == -1:
        slant= np.random.randint(-10,10) # generate random slant if no slant value is given

    rain_drops, drop_length= _generate_random_lines(imshape, slant, drop_length, rain_type)
    return _rain_process(tens, slant_extreme, drop_length, drop_color, drop_width, rain_drops)


def sim_fog(tens, fog_coeff=-1) -> Tensor:
    r"""
        Simulate foggy conditions on an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        fog_coeff (int): Coefficient value.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_fog(tens, fog_coeff=-1)
        >> filtered
        (427, 640, 3)

    """
    if fog_coeff != -1:
        if fog_coeff < 0.0 or fog_coeff > 1.0:
            raise ValueError('Fog coefficient must be between 0 and 1')

    imshape = tens.shape

    if fog_coeff == -1:
        fog_coeff_t = random.uniform(0.3,1)
    else:
        fog_coeff_t = fog_coeff

    hw = int(imshape[1]//3*fog_coeff_t)
    haze_list = _generate_random_blur_coordinates(imshape,hw)
    for haze_points in haze_list: 
        tens = _add_blur(tens, haze_points[0], haze_points[1], hw, fog_coeff_t) 

    tens = cv.blur(tens, (hw//10, hw//10))
    
    return _rgb(tens)


def sim_gravel(tens, rectangular_roi=(-1,-1,-1,-1), num_patches=8) -> Tensor:
    r"""
        Simulate gravelly conditions on an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        rectangular_roi (tuple): Rectanglar co-ordinates of the intended region of interest. Default: (-1,-1,-1,-1).
        num_patches (int): Number of patches to operate on.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_gravel(tens)
        >> filtered
        (427, 640, 3)

    """
    if is_tuple(rectangular_roi) and is_numeric_list_or_tuple(rectangular_roi) and len(rectangular_roi)==4:
        x1 = rectangular_roi[0]
        y1 = rectangular_roi[1]
        x2 = rectangular_roi[2]
        y2 = rectangular_roi[3]
    else:
        raise ValueError('Rectangular ROI dimensions are invalid.')

    if rectangular_roi == (-1,-1,-1,-1):
        if is_tensor(tens):
            x1 = 0
            y1 = int(tens.shape[0]*3/4)
            x2 = tens.shape[1]
            y2 = tens.shape[0]
        else:
            x1 = 0
            y1 = int(tens[0].shape[0]*3/4)
            x2 = tens[0].shape[1]
            y2 = tens[0].shape[0]

    elif x1 == -1 or y1 == -1 or x2 == -1 or y2 == -1 or x2 <= x1 or y2 <= y1:
        raise ValueError('Rectangular ROI dimensions are invalid.')

    return _gravel_process(tens, x1, x2, y1, y2, num_patches)


def sim_sun_flare(tens, flare_center=-1, angle=-1, num_flare_circles=8, src_radius=400, src_color=(255,255,255)) -> Tensor:
    r"""
        Add a source of light (flare) on an specific region of an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        flare_center (int): Center of the flare. Default: -1.
        angle (int): Angle of the flare. Default: -1
        num_flare_circles (int): Number of flare circles to operate on.
        src_radius (int): Intended size of the flare
        src_color (tuple): Intended source color of the flare
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_sun_flare(tens)
        >> filtered
        (427, 640, 3)

    """
    if angle != -1:
        angle = angle % (2*math.pi)

    if not(num_flare_circles >= 0 and num_flare_circles <= 20):
        raise ValueError('Numeric value must be between 0 and 20')

    imshape = tens.shape
    if angle == -1:
        angle_t = random.uniform(0, 2*math.pi)
        if angle_t == math.pi/2:
            angle_t = 0
    else:
        angle_t = angle

    if flare_center == -1:
        flare_center_t = (random.randint(0, imshape[1]), random.randint(0, imshape[0]//2))
    else:
        flare_center_t = flare_center

    x, y = _add_sun_flare_line(flare_center_t, angle_t, imshape)

    return _add_sun_process(tens, num_flare_circles, flare_center_t, src_radius, x, y, src_color)


def sim_motion_blur(tens, speed_coeff=-1) -> Tensor:
    r"""
        Simulate motion-blur conditions on an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        speed_coeff (int, float): Speed coefficient. Value must be between 0 and 1.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_motion_blur(tens, speed_coeff=-1)
        >> filtered
        (427, 640, 3)

    """
    if speed_coeff != -1:
        if speed_coeff < 0.0 or speed_coeff > 1.0:
            raise ValueError('Speed coefficient must be between 0 and 1')

    if speed_coeff == -1:
        count_t = int(15 * random.uniform(0, 1))
    else:
        count_t = int(15 * speed_coeff)

    return _apply_motion_blur(tens, count_t)


def sim_autumn(tens) -> Tensor:
    r"""
        Simulate autumn conditions on an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_autumn(tens)
        >> filtered
        (427, 640, 3)

    """
    return _autumn_process(tens)


## ROI:(top-left x1,y1, bottom-right x2,y2), shadow_dimension=no. of sides of polygon generated
def sim_shadow(tens, num_shadows=1, rectangular_roi=(-1,-1,-1,-1), shadow_dimension=5) -> Tensor: 
    r"""
        Simulate shadowy conditions on an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        num_shadows (int): Number of shadows to work with. Value must be between 1 and 10.
        rectangular_roi (tuple): Rectanglar co-ordinates of the intended region of interest. Default: (-1,-1,-1,-1).
        shadow_dimensions (int): Number of shadow dimensions. Value must be > 3. 
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise
        >> filtered = caer.filters.sim_shadow(tens)
        >> filtered
        (427, 640, 3)

    """
    if not(is_numeric(num_shadows) and num_shadows >= 1 and num_shadows <= 10):
        raise ValueError('Only 1-10 shadows can be introduced in an image')

    if not(is_numeric(shadow_dimension) and shadow_dimension >= 3 and shadow_dimension <= 10):
        raise ValueError('Polygons with dimensions < 3 don\'t exist and take time to plot')

    if is_tuple(rectangular_roi) and is_numeric_list_or_tuple(rectangular_roi) and len(rectangular_roi)==4:
        x1 = rectangular_roi[0]
        y1 = rectangular_roi[1]
        x2 = rectangular_roi[2]
        y2 = rectangular_roi[3]
    else:
        raise ValueError('Rectangular ROI dimensions are not valid')

    if rectangular_roi==(-1,-1,-1,-1):
        x1 = 0
        
        if(is_tensor(tens)):
            y1 = tens.shape[0] // 2
            x2 = tens.shape[1]
            y2 = tens.shape[0]
        else:
            y1 = tens[0].shape[0] // 2
            x2 = tens[0].shape[1]
            y2 = tens[0].shape[0]

    elif x1 == -1 or y1 == -1 or x2 == -1 or y2 == -1 or x2 <= x1 or y2 <= y1:
        raise ValueError('Rectangular ROI dimensions are not valid')

    return _shadow_process(tens,num_shadows, x1, y1, x2, y2, shadow_dimension)


#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import cv2 as cv

from ..coreten import Tensor, to_tensor
from ._constants import GRAY2BGR, GRAY2RGB
from ._bgr import bgr2lab, bgr2hsv, bgr2hls, bgr2yuv, bgr2luv

__all__ = [
    'gray2rgb',
    'gray2bgr',
    'gray2lab',
    'gray2hsv',
    'gray2hls',
    'gray2yuv',
    'gray2luv',
]


def _is_gray_image(tens):
    # tens = to_tensor(tens)
    # return tens.is_gray()
    return (len(tens.shape) == 2) or (len(tens.shape) == 3 and tens.shape[-1] == 1)


def gray2rgb(tens) -> Tensor:
    r'''
        Converts a Grayscale Tensor to its RGB version.

    Args:
        tens (Tensor): Valid Grayscale Tensor

    Returns:
        RGB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 2

    '''
    if not _is_gray_image(tens):
        raise ValueError(
            f'Tensor of shape 2 expected. Found shape {len(tens.shape)}. This function converts a Grayscale Tensor to its RGB counterpart'
        )

    im = cv.cvtColor(tens, GRAY2RGB)
    return to_tensor(im, cspace='rgb')


def gray2bgr(tens) -> Tensor:
    r'''
        Converts a Grayscale Tensor to its BGR version.

    Args:
        tens (Tensor): Valid Grayscale Tensor

    Returns:
        BGR Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 2

    '''
    if not _is_gray_image(tens):
        raise ValueError(
            f'Tensor of shape 2 expected. Found shape {len(tens.shape)}. This function converts a Grayscale Tensor to its BGR counterpart'
        )

    im = cv.cvtColor(tens, GRAY2BGR)
    return to_tensor(im, cspace='bgr')


def gray2hsv(tens) -> Tensor:
    r'''
        Converts a Grayscale Tensor to its HSV version.

    Args:
        tens (Tensor): Valid Grayscale Tensor

    Returns:
        HSV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 2

    '''
    if not _is_gray_image(tens):
        raise ValueError(
            f'Tensor of shape 2 expected. Found shape {len(tens.shape)}. This function converts a LAB Tensor to its HSV counterpart'
        )

    bgr = gray2bgr(tens)

    im = bgr2hsv(bgr)
    return to_tensor(im, cspace='hsv')


def gray2hls(tens) -> Tensor:
    r'''
        Converts a Grayscale Tensor to its HLS version.

    Args:
        tens (Tensor): Valid Grayscale Tensor

    Returns:
        HLS Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 2

    '''
    if not _is_gray_image(tens):
        raise ValueError(
            f'Tensor of shape 2 expected. Found shape {len(tens.shape)}. This function converts a LAB Tensor to its HLS counterpart'
        )

    bgr = gray2bgr(tens)

    im = bgr2hls(bgr)
    return to_tensor(im, cspace='hls')


def gray2lab(tens) -> Tensor:
    r'''
        Converts a Grayscale Tensor to its LAB version.

    Args:
        tens (Tensor): Valid Grayscale Tensor

    Returns:
        LAB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 2

    '''
    if not _is_gray_image(tens):
        raise ValueError(
            f'Tensor of shape 2 expected. Found shape {len(tens.shape)}. This function converts a Grayscale Tensor to its LAB counterpart'
        )

    bgr = gray2bgr(tens)

    im = bgr2lab(bgr)
    return to_tensor(im, cspace='lab')


def gray2yuv(tens) -> Tensor:
    r'''
        Converts a Grayscale Tensor to its YUV version.

    Args:
        tens (Tensor): Valid Grayscale Tensor

    Returns:
        YUV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 2

    '''
    if not _is_gray_image(tens):
        raise ValueError(
            f'Tensor of shape 2 expected. Found shape {len(tens.shape)}. This function converts a Grayscale Tensor to its YUV counterpart'
        )

    bgr = gray2bgr(tens)

    im = bgr2yuv(bgr)
    return to_tensor(im, cspace='yuv')


def gray2luv(tens) -> Tensor:
    r'''
        Converts a Grayscale Tensor to its LUV version.

    Args:
        tens (Tensor): Valid Grayscale Tensor

    Returns:
        LUV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 2

    '''
    if not _is_gray_image(tens):
        raise ValueError(
            f'Tensor of shape 2 expected. Found shape {len(tens.shape)}. This function converts a Grayscale Tensor to its LUV counterpart'
        )

    bgr = gray2bgr(tens)

    im = bgr2luv(bgr)
    return to_tensor(im, cspace='luv')



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

import numpy as np 
import cv2 as cv 
import random 
import collections

from ..coreten import Tensor, to_tensor
from .._internal import _check_target_size
from ..globals import (
    INTER_AREA, INTER_CUBIC, INTER_NEAREST, INTER_LINEAR
)

pad_to_str = {
    'constant':  0,
    'edge':      1,
    'reflect':   4,
    'symmetric': 2
}

MAX_VALUES_BY_DTYPE = {
    np.dtype("uint8"): 255,
    np.dtype("uint16"): 65535,
    np.dtype("uint32"): 4294967295,
    np.dtype("float32"): 1.0,
}

__all__ = [
    'hflip',
    'vflip',
    'hvflip',
    'rand_flip',
    'transpose',
    'scale',
    'rotate',
    'translate',
    'solarize',
    'posterize',
    'equalize',
    'clip',
    'pad'
]

def _is_rgb_image(tens):
    tens = to_tensor(tens, override_checks=True)
    return tens.is_rgb()
    # return len(tens.shape) == 3 and tens.shape[-1] == 3


def _is_gray_image(tens):
    tens = to_tensor(tens, override_checks=True)
    return tens.is_gray()
    # return (len(tens.shape) == 2) or (len(tens.shape) == 3 and tens.shape[-1] == 1)


def _get_num_channels(tens):
    return tens.shape[2] if len(tens.shape) == 3 else 1


def hflip(tens) -> Tensor:
    r"""
        Flip an image horizontally. 
    Args:
        tens (Tensor): Image to be flipped.

    Returns:
        Flipped image.

    """

    tens = np.ascontiguousarray(tens[:, ::-1, ...])
    return to_tensor(tens, override_checks=True)


def vflip(tens) -> Tensor:
    r"""
        Flip an image vertically. 
    Args:
        tens (Tensor): Image to be flipped.

    Returns:
        Flipped image.
        
    """
    tens = np.ascontiguousarray(tens[::-1, ...])
    return to_tensor(tens, override_checks=True)


def hvflip(tens) -> Tensor:
    r"""
        Flip an image both horizontally and vertically. 

    Args:
        tens (Tensor): Image to be flipped.

    Returns:
        Flipped image.
        
    """
    return hflip(vflip(tens))


def rand_flip(tens) -> Tensor: 
    r"""
        Randomly flip an image vertically or horizontally. 

    Args:
        tens (Tensor): Image to be flipped.

    Returns:
        Flipped image.
        
    """
    p = random.uniform(0, 1)

    if p > 0.5:
        return vflip(tens)
    else:
        return hflip(tens)


def transpose(tens) -> Tensor:
    if len(tens.shape) > 2:
        return tens.transpose(1, 0, 2)
    else:
        return tens.transpose(1, 0)


def rotate(tens, angle, rotPoint=None) -> Tensor:
    r"""
        Rotates an given image by an angle around a particular rotation point (if provided) or centre otherwise.
        
    """
    # h, w = image.shape[:2]
    # (cX, cY) = (w/2, h/2)

    # # Computing the sine and cosine (rotation components of the matrix)
    # transMat = cv.getRotationMatrix2D((cX, cY), angle, scale=1.0)
    # cos = np.abs(transMat[0, 0])
    # sin = np.abs(transMat[0, 1])

    # # compute the new bounding dimensions of the image
    # nW = int((h*sin) + (w*cos))
    # nH = int((h*cos) + (w*sin))

    # # Adjusts the rotation matrix to take into account translation
    # transMat[0, 2] += (nW/2) - cX
    # transMat[1, 2] += (nH/2) - cY

    # # Performs the actual rotation and returns the image
    # return cv.warpAffine(image, transMat, (nW, nH))

    height, width = tens.shape[:2]

    # If no rotPoint is specified, we assume the rotation point to be around the centre
    if rotPoint is None:
        rotPoint = (width//2, height//2)

    rotMat = cv.getRotationMatrix2D(rotPoint, angle, scale=1.0)

    tens = cv.warpAffine(tens, rotMat, (width, height))
    return to_tensor(tens, override_checks=True)


def translate(image, x, y) -> Tensor:
    r"""Translates a given image across the x-axis and the y-axis

    Args:
        x (int): shifts the image right (positive) or left (negative)
        y (int): shifts the image down (positive) or up (negative)
    
    Returns:
        The translated image

    """
    transMat = np.float32([[1, 0, x], [0, 1, y]])
    return cv.warpAffine(image, transMat, (image.shape[1], image.shape[0]))


def scale(tens, scale_factor, interpolation='bilinear') -> Tensor:
    interpolation_methods = {
        'nearest': INTER_NEAREST, '0': INTER_NEAREST, 0: INTER_NEAREST, # 0
        'bilinear': INTER_LINEAR, '1': INTER_LINEAR,  1: INTER_LINEAR,  # 1
        'bicubic': INTER_CUBIC,   '2': INTER_CUBIC,   2: INTER_CUBIC,   # 2
        'area': INTER_AREA,       '3': INTER_AREA,    3: INTER_AREA     # 3
    }
    if interpolation not in interpolation_methods:
        raise ValueError('Specify a valid interpolation type - area/nearest/bicubic/bilinear')

    if scale_factor > 1:
        # Neater, more precise
        interpolation = 'bicubic'

    height, width = tens.shape[:2]
    new_height, new_width = int(height * scale_factor), int(width * scale_factor)

    tens = cv.resize(tens, (new_width,new_height), interpolation=interpolation)
    return to_tensor(tens, override_checks=True)


def pad(tens, padding, fill=0, padding_mode='constant') -> Tensor:
    r"""
        Pad the given image on all sides with specified padding mode and fill value.

    Args:
        tens (Tensor): image to be padded.
        padding (int or tuple): Padding on each border. If a single int is provided this 
            is used to pad all borders. If tuple of length 2 is provided this is the padding
            on left/right and top/bottom respectively. If a tuple of length 4 is provided
            this is the padding for the left, top, right and bottom borders
            respectively.
        fill: Pixel fill value for constant fill. Default is 0. If a tuple of
            length 3, it is used to fill R, G, B channels respectively.
            This value is only used when the padding_mode is constant
        padding_mode: Type of padding. Should be: constant, edge, reflect or symmetric. Default is constant.
            - constant: pads with a constant value, this value is specified with fill
            - edge: pads with the last value on the edge of the image
            - reflect: pads with reflection of image (without repeating the last value on the edge)
                       padding [1, 2, 3, 4] with 2 elements on both sides in reflect mode
                       will result in [3, 2, 1, 2, 3, 4, 3, 2]
            - symmetric: pads with reflection of image (repeating the last value on the edge)
                         padding [1, 2, 3, 4] with 2 elements on both sides in symmetric mode
                         will result in [2, 1, 1, 2, 3, 4, 4, 3]
                         
    Returns:
        Tensor of shape ``(height, width, channels)``.

    """
    if not isinstance(padding, (tuple, list)):
        raise TypeError('Got inappropriate padding argument')

    if not isinstance(fill, (str, tuple)):
        raise TypeError('Got inappropriate fill argument')

    if not isinstance(padding_mode, str):
        raise TypeError('Got inappropriate padding_mode argument')

    if isinstance(padding, collections.Sequence) and len(padding) not in [2, 4]:
        raise ValueError(f'Padding must be an int or a 2, or 4 element tuple, not a {len(padding)} element tuple')

    assert padding_mode in ['constant', 'edge', 'reflect', 'symmetric'], \
        'Padding mode should be either constant, edge, reflect or symmetric'

    if isinstance(padding, int):
        pad_left = pad_right = pad_top = pad_bottom = padding

    if isinstance(padding, collections.Sequence) and len(padding) == 2:
        pad_left = pad_right = padding[0]
        pad_top = pad_bottom = padding[1]

    if isinstance(padding, collections.Sequence) and len(padding) == 4:
        pad_left = padding[0]
        pad_top = padding[1]
        pad_right = padding[2]
        pad_bottom = padding[3]


        tens = cv.copyMakeBorder(tens,
                                top = pad_top,
                                bottom = pad_bottom,
                                left = pad_left,
                                right = pad_right,
                                borderType = pad_to_str[padding_mode],
                                value = fill)

        return to_tensor(tens, override_checks=True)

                                
def crop(tens, x_min, y_min, x_max, y_max) -> Tensor:
    height, width = tens.shape[:2]
    if x_max <= x_min or y_max <= y_min:
        raise ValueError(
            "We should have x_min < x_max and y_min < y_max. But we got"
            " (x_min = {x_min}, y_min = {y_min}, x_max = {x_max}, y_max = {y_max})".format(
                x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max
            )
        )

    if x_min < 0 or x_max > width or y_min < 0 or y_max > height:
        raise ValueError(
            "Values for crop should be non negative and equal or smaller than image sizes"
            "(x_min = {x_min}, y_min = {y_min}, x_max = {x_max}, y_max = {y_max}, "
            "height = {height}, width = {width})".format(
                x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max, height=height, width=width
            )
        )

    return to_tensor(tens[y_min:y_max, x_min:x_max], override_checks=True)


def center_crop(image, target_size=None) -> Tensor:
    r"""Computes the centre crop of an image using `target_size`

    Args:
        image (Tensor): Valid image Tensor
        target_size (tuple): Size of the centre crop. Must be in the format `(width,height)`
    
    Returns:
        Cropped Centre (Tensor)
    
    Examples::

        >> tens = caer.data.bear() # Standard 640x427 image
        >> cropped = caer.center_crop(tens, target_size=(200,200))
        >> cropped.shape
        (200,200,3)

    """
    return _compute_centre_crop(image, target_size)


def rand_crop(tens, crop_height, crop_width, h_start, w_start) -> Tensor:
    height, width = tens.shape[:2]
    if height < crop_height or width < crop_width:
        raise ValueError(
            "Requested crop size ({crop_height}, {crop_width}) is "
            "larger than the image size ({height}, {width})".format(
                crop_height=crop_height, crop_width=crop_width, height=height, width=width
            )
        )
    x1, y1, x2, y2 = _get_random_crop_coords(height, width, crop_height, crop_width, h_start, w_start)

    return to_tensor(tens[y1:y2, x1:x2], override_checks=True)


def _compute_centre_crop(tens, target_size) -> Tensor:
    _ = _check_target_size(target_size)

    # Getting org height and target
    org_h, org_w = tens.shape[:2]
    target_w, target_h = target_size

    # The following line is actually the right way of accessing height and width of an opencv-specific image (height, width). However for some reason, while the code runs, this is flipped (it now becomes (width,height)). Testing needs to be done to catch this little bug
    # org_h, org_w = tens.shape[:2]


    if target_h > org_h or target_w > org_w:
        raise ValueError('To compute centre crop, target size dimensions must be <= tens dimensions')

    diff_h = (org_h - target_h) // 2
    diff_w = (org_w - target_w ) // 2
    
    # tens[y:y+h, x:x+h]
    return to_tensor(tens[diff_h:diff_h + target_h, diff_w:diff_w + target_w], override_checks=True)


def _get_random_crop_coords(height, width, crop_height, crop_width, h_start, w_start):
    y1 = int((height - crop_height) * h_start)
    y2 = y1 + crop_height
    x1 = int((width - crop_width) * w_start)
    x2 = x1 + crop_width
    return x1, y1, x2, y2


def solarize(tens, threshold=128) -> Tensor:
    r"""
        Invert all pixel values above a threshold.

    Args:
        tens (Tensor): The image to solarize.
        threshold (int): All pixels above this grayscale level are inverted.

    Returns:
        Solarized image (Tensor)
    
    Examples::

        >> tens = caer.data.sunrise()
        >> solarized = caer.solarize(tens, threshold=128)
        >> solarized.shape
        (427,640,3)

    """
    tens = to_tensor(tens, override_checks=True)
    max_val = MAX_VALUES_BY_DTYPE[tens.dtype]

    if tens.dtype == np.dtype("uint8"):
        lut = [(i if i < threshold else max_val - i) for i in range(max_val + 1)]

        prev_shape = tens.shape
        tens = cv.LUT(tens, np.array(lut, dtype=tens.dtype))

        if len(prev_shape) != len(tens.shape):
            tens = np.expand_dims(tens, -1)

        return to_tensor(tens, override_checks=True)

    result_tens = tens.copy()
    cond = tens >= threshold
    result_tens[cond] = max_val - result_tens[cond]
    return to_tensor(result_tens, override_checks=True)


def posterize(tens, bits) -> Tensor:
    r"""Reduce the number of bits for each color channel in the image.

    Args:
        tens (Tensor): Image to posterize.
        bits (int): Number of high bits. Must be in range [0, 8]

    Returns:
        Image with reduced color channels (Tensor)
    
    Examples::

        >> tens = caer.data.sunrise()
        >> posterized = caer.posterize(tens, bits=4)
        >> posterized.shape
        (427,640,3)

    """
    tens = to_tensor(tens, override_checks=True)
    bits = np.uint8(bits)

    if tens.dtype != np.uint8:
        raise TypeError("Image must have uint8 channel type")

    if np.any((bits < 0) | (bits > 8)):
        raise ValueError("bits must be in range [0, 8]")

    if not bits.shape or len(bits) == 1:
        if bits == 0:
            return np.zeros_like(tens)
        if bits == 8:
            return tens.copy()

        lut = np.arange(0, 256, dtype=np.uint8)
        mask = ~np.uint8(2 ** (8 - bits) - 1)
        lut &= mask

        return to_tensor(cv.LUT(tens, lut), override_checks=True)
        

    if not _is_rgb_image(tens):
        raise TypeError("If `bits` is iterable, image must be RGB")

    result_tens = np.empty_like(tens)
    for i, channel_bits in enumerate(bits):
        if channel_bits == 0:
            result_tens[..., i] = np.zeros_like(tens[..., i])
        elif channel_bits == 8:
            result_tens[..., i] = tens[..., i].copy()
        else:
            lut = np.arange(0, 256, dtype=np.uint8)
            mask = ~np.uint8(2 ** (8 - channel_bits) - 1)
            lut &= mask

            result_tens[..., i] = cv.LUT(tens[..., i], lut)

    return to_tensor(result_tens, override_checks=True)


def clip(tens, dtype, maxval) -> Tensor:
    tens = np.clip(tens, 0, maxval).astype(dtype)
    return to_tensor(tens, override_checks=True)


def _equalize_cv(tens, mask=None) -> Tensor:
    if mask is None:
        tens = cv.equalizeHist(tens)
        return to_tensor(tens, override_checks=True)

    histogram = cv.calcHist([tens], [0], mask, [256], (0, 256)).ravel()
    i = 0
    for val in histogram:
        if val > 0:
            break
        i += 1
    i = min(i, 255)

    total = np.sum(histogram)
    if histogram[i] == total:
        tens = np.full_like(tens, i)
        return to_tensor(tens, override_checks=True)

    scale = 255.0 / (total - histogram[i])
    _sum = 0

    lut = np.zeros(256, dtype=np.uint8)
    i += 1
    for i in range(i, len(histogram)):
        _sum += histogram[i]
        lut[i] = clip(round(_sum * scale), np.dtype("uint8"), 255)

    tens = cv.LUT(tens, lut)
    return to_tensor(tens, override_checks=True)


def equalize(tens, mask=None, by_channels=True) -> Tensor:
    r"""Equalize the image histogram.

    Args:
        tens (Tensor): RGB or grayscale image.
        mask (Tensor): An optional mask.  If given, only the pixels selected by the mask are included in the analysis. Maybe 1 channel or 3 channel array.
        by_channels (bool): If True, use equalization by channels separately, else convert image to YCbCr representation and use equalization by `Y` channel.

    Returns:
        Equalized image (Tensor)
    

    Examples::

        >> tens = caer.data.beverages()   
        >> equalized = caer.equalize(tens, mask=None)  
        >> equalized.shape   
        (427,640,3)
    
    """
    if tens.dtype != np.uint8:
        raise TypeError("Image must have uint8 channel type")

    if mask is not None:
        if _is_rgb_image(mask) and _is_gray_image(tens):
            raise ValueError("Wrong mask shape. Image shape: {}. Mask shape: {}".format(tens.shape, mask.shape))

        if not by_channels and not _is_gray_image(mask):
            raise ValueError("When `by_channels=False`, only 1-channel mask is supported. Mask shape: {}".format(mask.shape))

    if mask is not None:
        mask = mask.astype(np.uint8)

    if _is_gray_image(tens): 
        return to_tensor(_equalize_cv(tens, mask), override_checks=True)

    if not by_channels:
        result_tens = cv.cvtColor(tens, cv.COLOR_RGB2YCrCb)
        result_tens[..., 0] = _equalize_cv(result_tens[..., 0], mask)
        tens = cv.cvtColor(result_tens, cv.COLOR_YCrCb2RGB) 
        return to_tensor(tens, override_checks=True)

    result_tens = np.empty_like(tens)
    for i in range(3):
        if mask is None:
            _mask = None
        elif _is_gray_image(mask):
            _mask = mask
        else:
            _mask = mask[..., i]

        result_tens[..., i] = _equalize_cv(tens[..., i], _mask)

    return to_tensor(result_tens, override_checks=True)


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


from threading import Thread
import time
import math
from queue import Queue
import cv2 as cv
import numpy as np 

from .constants import (
    FPS, FRAME_COUNT, FRAME_HEIGHT, FRAME_WIDTH
)
from ..annotations import Tuple
from ..coreten import Tensor

__all__ = [
    'Stream'
]


#pylint:disable=no-member

# ret, jpeg = cv2.imencode('.jpg', image)
# return jpeg.tobytes()

# This class can handle both live as well as pre-existing videos. 
class Stream:
    r"""
        This is an auxiliary class that enables Video Streaming for ``caer`` with minimalistic latency, and at the expense
        of little to no additional computational requirements.
        
        The basic idea behind it is to tracks and save the salient feature array for the given number of frames and then uses these anchor point to cancel out all perturbations relative to it for the incoming frames in the queue. This class relies heavily on **Threaded Queue mode** for error-free & ultra-fast frame handling.

    Args:
        source (int, str): Source path for the video. Uses an external camera device if ``source`` is an integer.
        qsize (int): Default queue size for handling the video streams. Default: 128.
    """

    def __init__(self, source = 0, qsize=128) -> None: # TODO: Add colorspace support
        r"""
            Source must either be an integer (0,1,2 etc) or a path to a video file
        """
        self.live_video = False
        
        if isinstance(source, int):
            self.live_video = True
            # raise ValueError('Expected a filepath. Got an integer. FileVideoStream is not for live feed. Use LiveVideoStream instead')

        if not isinstance(source, (int,str)):
            raise ValueError(f'Expected either an integer or filepath. Got {type(source)}')
        
        # initializing the video stream
        self._video_stream = cv.VideoCapture(source)
        self.kill_stream = False

        self.width = int(self._video_stream.get(FRAME_WIDTH))
        self.height = int(self._video_stream.get(FRAME_HEIGHT))
        self.res = (self.width, self.height)

        self.fps = math.ceil(self._video_stream.get(FPS))
        self.frames = int(self._video_stream.get(FRAME_COUNT))
        
        # initialize the queue to store frames 
        self._Q = Queue(maxsize=qsize)

        # intialize thread
        self._thread = None


    def start(self) -> None:
        # start a thread to read frames from the video stream
        self._thread = Thread(target=self._update, name="caer.video.Stream()", args=())
        self._thread.daemon = True
        self._thread.start()
        return self


    def _update(self) -> Tensor:
        while True:
            if self.kill_stream:
                break

            # otherwise, ensure the queue has room in it
            if not self._Q.full():
                # read the next frame from the file
                ret, frame = self._video_stream.read()

                # If at the end of the video stream
                if not ret:
                    self.release()
                    return 

                # add the frame to the queue
                self._Q.put(frame)
            else:
                time.sleep(0.1)  # Rest for 10ms if we have a full queue

        self._video_stream.release()


    def read(self) -> Tensor:
        """
        Extracts frames synchronously from monitored deque, while maintaining a fixed-length frame buffer in the memory, and blocks the thread if the deque is full.

        **Returns:** A n-dimensional numpy array.
        """

        return self._Q.get()


    def release(self) -> None:
        """
        Safely terminates the thread, and release the Stream resources.
        """
        self.kill_stream = True
        # wait until stream resources are released
        if self._thread is not None:
            self._thread.join()
            self._thread = None 
        self.frames = 0 
        self.fps = 0
    

    # Gets frame count
    def count_frames(self) -> int:
        """
            Returns the number of frames for the current video
            Optional: use the `frames` attribute
        """
        if not self.kill_stream and not self.live_video:
            return self.frames
            # if get_opencv_version() == '2':
            #     return int(self.stream.get(FRAME_COUNT_DEPR))
            # else:
            #     return int(self.stream.get(FRAME_COUNT))
            

        if self.live_video:
            print('[WARNING] Frames cannot be computed on live streams')
            return -1


    # Gets FPS count
    def get_fps(self) -> (float, int):
        """
            Returns the fps (frames per second) value for the current video
            Optional: use the `fps` attribute
        """
        if not self.kill_stream:
            return self.fps
            # if get_opencv_version() == '2':
            #     return math.ceil(self.stream.get(FPS_DEPR))
            # else:
            #     return math.ceil(self.stream.get(FPS))
    
    
    # Get frame dimensions
    def get_res(self) -> Tuple[int]:
        return self.res



###########################################################################################
# Old implementation of stream.py when it used to inherit the FileStream() from filestream.py
# Until 5 Dec 2020, this was the implementation
###########################################################################################

# from .filestream import FileStream


# __all__ = [
#     'Stream'
# ]

# # Using the FileStream class as it can handle both live as well as pre-existing videos
# class Stream(FileStream):
#     """
#     Stream() supports a diverse range of video streams which can handle/control video stream almost any IP/USB Cameras, multimedia video file format (tested upto 4k), any network stream URL such as *http(s), rtp, rstp, rtmp, mms, etc.*

#     It provides a flexible, high-level multi-threaded wrapper around OpenCV's VideoCapture() for threaded, error-free and synchronized frame handling.
#     """

#     def __init__(self, source=0):
#         # Initializing the stream from DefaultVideoStream
#         print('Beg of f')
#         self._video_stream = FileStream(source=source)
#         print('End of f')

#         self.width = self._video_stream.width
#         self.height = self._video_stream.height
#         self.res = (self.width, self.height)

#         self.fps = self._video_stream.fps
#         self.frames = self._video_stream.frames

#     def start(self):
#         # Begins the threaded video stream
#         return self._video_stream.start()
    
#     def update(self):
#         self._video_stream.update()
      
#     def read(self):
#         """
#         Extracts frames synchronously from monitored deque, while maintaining a fixed-length frame buffer in the memory, and blocks the thread if the deque is full.

#         **Returns:** A n-dimensional numpy array.
#         """
        
#         return self._video_stream.read()
    
#     def count_frames(self):
#         return self._video_stream.count_frames()
        
#     def release(self):
#         """
#         Safely terminates the thread, and release the VideoStream resources.
#         """
#         self._video_stream.release()
        

#      # Get FPS
#     def get_fps(self):
#         return self._video_stream.get_fps()

#     # Get frame dimensions
#     def get_res(self):
#         return self._video_stream.get_res()



#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

from .core import (
    to_bgr,
    to_rgb,
    to_gray,
    to_hsv,
    to_hls,
    to_lab,
    to_yuv,
    to_luv,
    __all__ as __all_core__,
)

from ._bgr import (
    bgr2rgb,
    bgr2gray,
    bgr2hsv,
    bgr2hls,
    bgr2lab,
    bgr2yuv,
    bgr2luv,
    __all__ as __all_bgr__,
)

from ._rgb import (
    rgb2bgr,
    rgb2gray,
    rgb2hsv,
    rgb2hls,
    rgb2lab,
    rgb2yuv,
    rgb2luv,
    __all__ as __all_rgb__,
)

from ._gray import (
    gray2rgb,
    gray2bgr,
    gray2hsv,
    gray2hls,
    gray2lab,
    gray2yuv,
    gray2luv,
    __all__ as __all_gray__,
)

from ._hsv import (
    hsv2rgb,
    hsv2bgr,
    hsv2gray,
    hsv2hls,
    hsv2lab,
    hsv2yuv,
    hsv2luv,
    __all__ as __all_hsv__,
)

from ._hls import (
    hls2rgb,
    hls2bgr,
    hls2gray,
    hls2hsv,
    hls2lab,
    hls2yuv,
    hls2luv,
    __all__ as __all_hls__,
)

from ._lab import (
    lab2rgb,
    lab2bgr,
    lab2gray,
    lab2hsv,
    lab2hls,
    lab2yuv,
    lab2luv,
    __all__ as __all_lab__,
)

from ._yuv import (
    yuv2rgb,
    yuv2bgr,
    yuv2gray,
    yuv2hsv,
    yuv2hls,
    yuv2lab,
    yuv2luv,
    __all__ as __all_yuv__,
)

from ._luv import (
    luv2rgb,
    luv2bgr,
    luv2gray,
    luv2hsv,
    luv2hls,
    luv2lab,
    luv2yuv,
    __all__ as __all_luv__,
)

# from .constants import (
#     IMREAD_COLOR,
#     BGR2RGB,
#     BGR2GRAY,
#     BGR2HSV,
#     RGB2GRAY,
#     RGB2BGR,
#     RGB2HSV,
#     BGR2LAB,
#     RGB2LAB,
#     HSV2BGR,
#     HSV2RGB,
#     LAB2BGR,
#     LAB2RGB,
#     GRAY2BGR,
#     GRAY2RGB,
#     HLS2BGR,
#     HLS2RGB,
#     __all__ as __all_const__
# )


# __all__ = __all_const__ + __all_core__ + __all_rgb__ + __all_hls__+ __all_gray__ + __all_bgr__ + __all_hsv__ + __all_lab__
__all__ = (
    __all_core__
    + __all_rgb__
    + __all_hls__
    + __all_gray__
    + __all_bgr__
    + __all_hsv__
    + __all_lab__
    + __all_yuv__
    + __all_luv__
)

# # Don't pollute namespace
# del __all_const__
del __all_bgr__
del __all_rgb__
del __all_gray__
del __all_hsv__
del __all_lab__
del __all_hls__
del __all_yuv__
del __all_luv__
